# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import os
import re
import xml.etree.ElementTree as ET

import reschema
ElementBase = ET.Element


class Document(object):
    def __init__(self, title, printable=False):
        self.printable = printable
        self.html = HTMLElement('html')
        self.head = self.html.head()
        self.head.title().text = title
        self.head.script(type="text/javascript", text="$script")
        self.head.style(type="text/css", text="$css")
        self.body = self.html.body(onload="set_favicon()")

        self.header = self.body.div(cls="header")
        if self.printable:
            self.header.attrib['style'] = 'display: none'

        if self.printable:
            self.main = self.body.div(cls="main-printable")
        else:
            self.main = self.body.div(cls="main")

        self.navbar = self.main.div(cls="navbar")
        if self.printable:
            self.navbar.attrib['style'] = 'display: none'
        self.toc = self.navbar.div(cls="toc")

        if self.printable:
            self.content = self.main.div(cls="content-printable")
        else:
            self.content = self.main.div(cls="content")

        self.menu = Menu(self.toc)

        self.footer = self.body.div(cls="footer")
        if self.printable:
            self.footer.attrib['style'] = 'display: none'

    def write(self, file):
        f = open(file, "w")
        text = ET.tostring(self.html, method="html")
        path = os.path.dirname(os.path.abspath(reschema.__file__))

        f2 = open(path + "/servicedef.css", "r")
        text = text.replace("$css", "<!--" + f2.read() + "-->")
        f2.close()

        f2 = open(path + "/servicedef.js", "r")
        text = text.replace("$script", f2.read())
        f2.close()

        text = text.replace("><pre", ">\n<pre")
        text = text.replace("</pre>", "</pre>\n")

        inpre = False
        for line in text.split("\n"):
            if inpre:
                if re.search("</pre>", line):
                    inpre = False
            else:
                if re.match("<pre", line):
                    if not re.search("</pre", line):
                        inpre = True
                else:
                    # Insert newline breaks before certain tags (but not all)
                    line = re.sub("><(div|p|ul|li|script|style|td|tr|th|table|col|colgroup)",
                                  ">\n<\\1",
                                  line)
            f.write(line + "\n")

        f.close()


class TabBar(object):
    def __init__(self, div, baseid, printable=False):
        self.div = div
        self.baseid_init = baseid
        self.baseid = baseid
        self.selected = True
        self.init = True
        self.count = 0
        self.printable = printable

    def init_tabbar(self):
        if self.printable:
            if self.count > 0:
                self.finish()
            self.baseid = self.baseid_init + str(self.count)
            self.count = self.count + 1
            self.selected = True
        self.tabContainer = self.div.div().div(cls='tabContainer', id='%s-tabs' % self.baseid)
        self.ul = self.tabContainer.div(cls='digiTabs', id='%s-tabbar' % self.baseid)
        self.init = False

    def add_tab(self, label, id, content):
        if self.init or self.printable:
            self.init_tabbar()

        self.ul.li(cls=('selected' if self.selected else None),
                   id='%s-tab-%s' % (self.baseid, id),
                   onclick='showtab("%s", "%s")' % (self.baseid, id)).text = label

        if self.selected:
            self.tabContainer.div(cls='tabContent', id='%s-tabcontent' % self.baseid).append(content)

        self.tabContainer.div(id='%s-%s' % (self.baseid, id), style='display:none').append(content)

        self.selected = False

    def finish(self):
        self.div.div(style="clear:both")


class HTMLElement(ElementBase):
    def __init__(self, name, *args, **kwargs):
        delargs = []
        for arg, value in kwargs.iteritems():
            if value is None:
                delargs.append(arg)
        for arg in delargs:
            del kwargs[arg]

        if 'cls' in kwargs:
            kwargs['class'] = kwargs['cls']
            del kwargs['cls']
        if 'text' in kwargs:
            self.text = kwargs['text']
            del kwargs['text']
        ElementBase.__init__(self, name, attrib=kwargs)

    def settext(self, *lst):
        last = None
        newlist = []

        # Need to first join up any strings in a row
        for e in lst:
            if last is not None:
                if (((type(last) is str) or (type(last) is unicode)) and
                    ((type(e) is str) or (type(e) is unicode))):
                    last += e
                else:
                    newlist.append(last)
                    last = e
            else:
                last = e
        newlist.append(last)

        last = None
        for e in newlist:
            if (type(e) is str) or (type(e) is unicode):
                if last is None:
                    self.text = e
                else:
                    last.tail = e
            else:
                last = e

    def tostring(self):
        return ET.tostring(self)

    def subelement(self, name, *args, **kwargs):
        sube = HTMLElement(name, **kwargs)
        self.append(sube)
        return sube

    def __getattr__(self, name):
        def method(*args, **kwargs):
            return self.subelement(name, *args, **kwargs)
        return method

    def table(self, *args, **kwargs):
        table = HTMLTable(*args, **kwargs)
        self.append(table)
        return table


class HTMLTable(HTMLElement):

    def __init__(self, *args, **kwargs):
        HTMLElement.__init__(self, 'table', *args, **kwargs)
        self._body = None

    def define_columns(self, classes):
        grp = self.colgroup()
        for _cls in classes:
            grp.col(cls=_cls)

    def row(self, cells, header=False):
        if self._body is None:
            self._body = self.tbody()

        tr = self._body.tr()
        tds = []
        for cell in cells:
            td = tr.th() if header else tr.td()
            tds.append(td)
            td.text = cell
        return tds


class Menu(HTMLElement):

    def __init__(self, parent, *args, **kwargs):
        HTMLElement.__init__(self, 'ul', *args, **kwargs)
        parent.append(self)

    def add_item(self, text, href=None):
        li = self.li()
        if href is not None:
            if isinstance(href, HTMLElement):
                href = '#' + href.attrib['id']
            li.a(href=href, text=text)

        else:
            li.text = text
        return li

    def add_submenu(self):
        return Menu(self)



# Copyright (c) 2014 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

"""
This module contains code for interacting with git repositories. It is
used to determine an appropriate version number from either the local
git repository tags or from a version file.
"""

from __future__ import unicode_literals, print_function, division
import os
import inspect
import re
from subprocess import Popen, PIPE
from collections import namedtuple


ALPHA_NUM_PERIOD_UNDER_HYPHEN = r'[\w._-]+$'
ALPHA_PERIOD_PLUS_UNDER_HYPHEN = r'[A-Za-z.+_-]+'


FINAL = '\d+(?:\.\d+)*'
PRE = '(a|b|r?c)\d+'
POST = '\.post\d+'
DEV = '\.dev\d+'
PEP440 = '{0}(?:{1})?(?:{2})?(?:{3})?\s*$'.format(FINAL, PRE, POST, DEV)
PEP8 = r'[a-z_]+'


class InvalidString(Exception):
    """Invalid strings"""
    def __init__(self, error):
        self.error = error


class InvalidTag(InvalidString):
    """Exception class for invalid tags"""
    def __str__(self):
        return "Invalid tag: {0}".format(self.error)


class InvalidBranch(InvalidString):
    """Exception class for invalid tags"""
    def __str__(self):
        return "Invalid branch {0}".format(self.error)


class InvalidCommand(InvalidString):
    """Exception class for invalid git command"""
    def __str__(self):
        return ("Invalid command {0}, \n Exception {1}"
                .format(' '.join(self.error[0]), self.error[1]))


def git(cmd, dir=None, input=False):
    """Return a git command results. raise an EnvironmentError if not a git repo.

    :param cmd: a list of strings with 'git' in front form a git command
    :param dir: dir to change to before run the git command
    :param input: flag indicating whether the git command has man-input params.
                  If yes, stderr would mean the param is invalid
    """
    cwd = None
    try:
        if dir is not None:
            cwd = os.getcwd()
            os.chdir(dir)
        process = Popen(['git'] + cmd, stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()
    finally:
        if cwd is not None:
            os.chdir(cwd)

    if stderr:
        if not input:
            # not a git repo
            raise EnvironmentError(stderr)
        else:
            raise InvalidCommand((['git'] + cmd, stderr))
    else:
        return stdout.strip()


def call_git_describe(abbrev=None):
    """Return 'git describe' output.

    :param abbrev: Integer to use as the --abbrev value for git describe
    """
    cmd = ['describe']

    # --abbrev and --long are mutually exclusive 'git describe' options
    if abbrev is not None:
        cmd.append('--abbrev={0}'.format(abbrev))

    return git(cmd)


def get_branch():
    """Return the current branch's name."""

    input = git(['branch'])
    line = [ln for ln in input.split('\n') if ln.startswith('*')][0]
    return line.split()[-1]


def get_parents(branch):
    """ Return the parent branch

    :param branch: name of the current branch
    """
    input = git(['show-branch'])
    lines = [ln for ln in input.split('\n') if '*' in ln]
    # line is in the form of '+ *++ [parent] commit message'
    res = set([])
    for ln in lines:
        name = ln.split('[')[1].split(']')[0]
        # stripping ^~ and integer on the right
        name = re.split('\^|~', name)[0]
        if name != branch:
            res.add(name)
    return res


def valid_pep440(string):
    """return True if string is a valid pep440 version

    :param string: version string
    """
    return True if re.match(PEP440, string) else False


def valid_public_ver(tag, pkg_name=None):
    """Return True if tag is a valid public PEP440 version

    if tag is prefixed by a package name, then the package name
    must match

    :param tag: tag string
    :param pkg_name: package name if not None, need to prefix tag
    """
    if pkg_name is None:
        return valid_pep440(tag)
    else:
        return (tag.startswith(pkg_name + '-') and
                valid_pep440(tag[len(pkg_name) + 1:]))


def tag2cmt(tag):
    """Return the commit that the tag points to

    :param tag: the tag on the inquired commit, type:string
    """
    input = git(['rev-list', '{0}'.format(tag)], input=True)
    return input.split('\n')[0]


def get_commit():
    """Return the latest commit"""
    cmt = git(['log', '-n', '1', "--pretty=format:'%H'"])
    # cmt is double-quoted string as "'XXXXXXXXXXX'"
    # need to be stripped one pair of quotes to be normal
    return cmt.replace("'", "")


def tag2branches(tag):
    """Return the branches containing the given tag in their history

    except the current branch.

    :param tag: the tag name
    """
    lines = git(['branch', '--contains', '{0}'.format(tag2cmt(tag))])

    # lines renders as : child\n  grand-child\n* grand-grand-child\n  master
    return set([ln.strip() for ln in lines.split('\n') if '*' not in ln])


def find_tag(pkg_name, nondev=False):
    """Return tag that can be used as a version using the following algorithm.

    the latest tag is not necessarily the one to use

    if the tag does not start with a pkg_name:
       if has not seen a tag with pkg_name,
            assume single-pkg repo, the tag is chosen
       else:
            assume multi-pkg repo, tag is wrong, error
    else:
       if matches pkg_name,
            the tag is chosen
       else:
            get next tag, go to the start
    if run out of commit history,
       error

    :param pkg_name: package name used to get a valid tag
    """

    lines = git(['for-each-ref', '--sort=taggerdate',
                '--format', "'%(refname) %(taggerdate)'", 'refs/tags'])
    lines = lines.split('\n')
    # the lines are ordered by time ascending
    seen_tag_with_pkg = False
    for ind in range(len(lines) - 1, -1, -1):
        # one tag is contained within a line as below:
        # "'refs/tags/<tag> Fri Sep 26 09:59:53 2014 -0400'"
        tag = lines[ind].replace("'", "").split(' ')[0][10:]

        if pkg_name is None:
            if re.match(PEP440, tag):
                if not nondev or ('.dev' not in tag):
                    return tag
        # pkg_name is not None
        elif re.match(PEP8 + '-\d.*$', tag):
            seen_tag_with_pkg = True
            if (tag.startswith(pkg_name + '-') and
                    re.match(PEP440, tag[len(pkg_name) + 1:])):
                if not nondev or ('.dev' not in tag):
                    return tag
        # tag does not start with a pkg_name
        elif seen_tag_with_pkg:
            raise InvalidTag(tag)
    raise InvalidTag("could not find a valid tag with pkg_name '{0}'"
                     .format(pkg_name))


def get_commits(tag):
    """Return the number of commits since the tag.

    :param tag: the tag name
    """
    try:
        return (int(git(['rev-list', '--count', 'HEAD'])) -
                int(git(['rev-list', '--count', '{0}'.format(tag)],
                        input=True)))
    except EnvironmentError:
        # Older git versions do not support '--count'
        return (len(git(['rev-list', 'HEAD']).split('\n')) -
                len(git(['rev-list', '{0}'.format(tag)],
                        input=True).split('\n')))


def git_info():
    """Return an git_info object, which contains attributes:
        branch: branch's name, type: string
        tag: most recent tag, type: string
        tagged_cmt: the commit that the most recent tag points to, type: string
        cmt: latest commit of the current branch, type:string
    """
    branch = get_branch()
    base_tag = call_git_describe(abbrev=0)
    GitInfo = namedtuple('GitInfo', 'branch tag tagged_cmt cmt')
    return GitInfo(branch=branch,
                   tag=base_tag,
                   tagged_cmt=tag2cmt(base_tag),
                   cmt=get_commit())


def verify_repository(pkg_file):
    """Raise an error if this source file is not in tracked by git.

    :param pkg_file: pkig_file to be tested
    """
    dirname = os.path.dirname(pkg_file)
    basename = os.path.basename(pkg_file)

    git(['ls-files', basename, '--error-unmatch'], dir=dirname)
    return


def valid_local_ver(version):
    """check if a version is valid.

    the version needs to consist of ASCII numbers, letters and periods
    :param version: version string to be validated
    """
    return True if re.match(ALPHA_NUM_PERIOD_UNDER_HYPHEN, version) else False


def increment_rightmost(version, number):
    """ increment the rightmost number of the version by the number

    :param version: the version string
    :param number: the number by which the rightmost number is incremented
    """
    num_str = re.split(ALPHA_PERIOD_PLUS_UNDER_HYPHEN, version)[-1]
    return version[:-len(num_str)] + str(int(num_str) + number)


def get_version(pkg_name=None, pkg_file=None, v_file='RELEASE-VERSION'):

    """Return PEP440 style version string.

    :param pkg_name: package name defaults to None for one-pkg repo
       if multi-pkg repo, package name refers to the pkg based on which
       release version will be generated

    :param pkg_file: Some filename in the package, used to test if this
       is a live git repostitory (defaults to caller's file)

    :param v_file: Fallback path name to a file where release_version is saved
    """
    if pkg_file is None:
        parent_frame = inspect.stack()[1]
        pkg_file = inspect.getabsfile(parent_frame[0])

    try:
        verify_repository(pkg_file)

        info = git_info()

        if (valid_public_ver(info.tag, pkg_name) and
                info.cmt == info.tagged_cmt):
            version = info.tag
        else:
            tag = find_tag(pkg_name)
            if (info.branch != 'master' and
                    len(tag2branches(tag).
                        intersection(get_parents(info.branch))) > 0):
                # the most recent tag is on a parent branch
                if valid_local_ver(info.branch):
                    # current branch name is a valid local version
                    non_dev_tag = find_tag(pkg_name, nondev=True)
                    commits = get_commits(non_dev_tag)
                    # get the hash string of the commit
                    sha = info.cmt[:7]
                    version = ('{0}+git.{1}.{2}.{3}'
                               .format(non_dev_tag, info.branch,
                                       commits, sha))
                else:
                    raise InvalidBranch("'{0}' is not a valid local version"
                                        .format(info.branch))

            elif '.dev' in tag:
                # recent tag is a development tag
                # increment the N in devN by number of commits
                version = increment_rightmost(tag, get_commits(tag))

            else:
                # recent tagged commit is non-development, increment the least
                # significant number in the string and append .devN where N
                # is the number of commits since the tag
                version = ('{0}.dev{1}'
                           .format(increment_rightmost(tag, 1),
                                   get_commits(tag)))

        with open(v_file, 'w') as f:
            f.write(version)

    except (EnvironmentError, InvalidString) as e:
        # Not a git repository or invalid tag
        # so fall back to reading RELEASE-VERSION
        if type(e) in [InvalidTag, InvalidCommand, InvalidBranch]:
            print (e)
        else:
            print ('Not a git repository')
        print ("Reading the version from file '{0}'".format(v_file))
        if (os.path.exists(v_file)):
            with open(v_file, 'r') as f:
                version = f.read().strip()
        else:
            version = 'unknown'

    return version



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from __future__ import print_function

import re
import uritemplate
import traceback
import reschema.exceptions
import reschema.settings
from reschema import jsonschema

INDENT_OFFSET = 4

reschema.settings.LOAD_DESCRIPTIONS = True
reschema.settings.MARKED_LOAD = True


def get_first_marked_object(*objects):
    """ Given a list of objects, return the first that has marks.
    :param objects: A number of objects to check for.
    :returns: The first member of objects with a "start_mark" attribute or None

    """
    for obj in objects:
        if hasattr(obj, "start_mark"):
            return obj


class ValidationFail(reschema.exceptions.MarkedError):
    """
    Schema did something it shouldn't have
    """
    pass


class RuleDisabled(Exception):
    """
    Rule is disabled due to a relint-disable tag
    """
    pass


class Rule(object):
    """
    Base class for a rule validator.  Exists mainly to isolate the validation
    callback from the code which checks for relint-disable tags in the
    servicedef.
    """

    def __init__(self, rule_id, rule_func):
        """
        :param rule_id: Rule's Id, for disabling/reporting purposes
        :param rule_func: Rule validation callback
        """

        self.rule_id = rule_id
        self._rule_func = rule_func

    def __call__(self, schema, disabled):
        """
        Run the rule

        :param schema: schema to verify
        :type: reschema.jsonschema.Schema or derived
               reschema.jsonschema.Link or
               reschema.jsonschema.Relation or
               reschema.servicedef.ServiceDef

        :param disabled: list of disabled rules
        """

        if self.rule_id in disabled:
            raise RuleDisabled

        self._rule_func(schema)


class Result(object):
    """ Helper to hold the result of a rule execution """

    PASSED = 0
    FAILED = 1
    DISABLED = 2

    def __init__(self, rule_id, obj_id, status, message='', exc=None):
        """
        :param rule_id: Rule that executed
        :param obj_id: Object checked (in json-pointer format)
        :param status: PASSED/FAILED/DISABLED
        :param message: Optional additional information
        :param exc: Optional exception that was raised to cause this Result.

        """

        self.rule_id = rule_id
        self.obj_id = obj_id
        self.status = status
        self.message = message
        self.exc = exc

    def __str__(self):
        status = ['PASS', 'FAIL', 'DIS '][self.status]

        # marks are easier to find in the document than the ID.
        if self.exc and self.exc.start_mark:
            return ('{name}:{line}:{column}: {status} [{rule}]  '
                    '{message}').format(name=self.exc.start_mark.name,
                                        line=self.exc.start_mark.line + 1,
                                        column=self.exc.start_mark.column + 1,
                                        status=status,
                                        rule=self.rule_id,
                                        message=self.message)
        return '{0}: [{1}] - \'{2}\' {3}'.format(status, self.rule_id,
                                                 self.obj_id, self.message)


class Validator(object):
    """
    Class which collects rules and executes them, tracking pass/fail/etc
    """

    SERVICEDEF_RULES = []
    TYPE_RULES = []
    RESOURCE_RULES = []
    LINK_RULES = []
    SCHEMA_RULES = []
    VERBOSITY = 0
    RVBD_BASE_URL = 'http://support.riverbed.com/apis'
    STD_LINKS = ['self', 'get', 'set', 'delete', 'create']

    @classmethod
    def servicedef(cls, rule_id):
        """ Registers a rule which operates on the servicedef """
        def wrapper(rule_func):
            rule = Rule(rule_id, rule_func)
            cls.SERVICEDEF_RULES.append(rule)

            return rule
        return wrapper

    @classmethod
    def typedef(cls, rule_id):
        """ Registers a rule which operates on a type """
        def wrapper(rule_func):
            rule = Rule(rule_id, rule_func)
            cls.TYPE_RULES.append(rule)

            return rule
        return wrapper

    @classmethod
    def resource(cls, rule_id):
        """ Registers a rule which operates on a resource """
        def wrapper(rule_func):
            rule = Rule(rule_id, rule_func)
            cls.RESOURCE_RULES.append(rule)

            return rule
        return wrapper

    @classmethod
    def link(cls, rule_id):
        """ Registers a rule which operates on a link """
        def wrapper(rule_func):
            rule = Rule(rule_id, rule_func)
            cls.LINK_RULES.append(rule)

            return rule
        return wrapper

    @classmethod
    def schema(cls, rule_id):
        def wrapper(rule_func):
            rule = Rule(rule_id, rule_func)
            cls.SCHEMA_RULES.append(rule)

            return rule
        return wrapper

    @classmethod
    def set_verbosity(cls, level):
        cls.VERBOSITY = level

    def check_schema(self, schema, results, disabled_rules_above):
        """
        Report any failures defined within a schema, including
        top-level failures as well as any failures in each subschema
        """
        if type(schema) is not jsonschema.Ref:
            rules = get_disabled(schema)
            disabled_rules = disabled_rules_above.union(rules)

            results.extend(self._run_rules(Validator.TYPE_RULES +
                                           Validator.SCHEMA_RULES,
                                           schema, disabled_rules))

            self.check_sub_schema(schema, results, disabled_rules)

    def check_sub_schema(self, schema, results, disabled_rules):
        """
        Check if any subschema defined within schema is valid.
        In particular, find any possible existing keywords and
        check individual subschema defined underneath the keyword.
        """
        if hasattr(schema, 'properties'):
            for _, prop in schema.properties.iteritems():
                self.check_schema(prop, results, disabled_rules)

        if hasattr(schema, 'items'):
            self.check_schema(schema.items, results, disabled_rules)

        if hasattr(schema, '_request') and schema._request is not None:
            if type(schema._request) not in [jsonschema.Ref, jsonschema.Null]:
                self.check_schema(schema._request, results, disabled_rules)

        if hasattr(schema, '_response') and schema._response is not None:
            if type(schema._response) not in [jsonschema.Ref, jsonschema.Null]:
                self.check_schema(schema._response, results, disabled_rules)

        if hasattr(schema, 'anyof'):
            for subschema in schema.anyof:
                self.check_schema(subschema, results, disabled_rules)

        if hasattr(schema, 'oneof'):
            for subschema in schema.oneof:
                self.check_schema(subschema, results, disabled_rules)

        if hasattr(schema, 'allof'):
            for subschema in schema.allof:
                self.check_schema(subschema, results, disabled_rules)

        if hasattr(schema, 'not_') and schema.not_ is not None:
            self.check_schema(schema.not_, results, disabled_rules)

        if hasattr(schema, '_params'):
            for _, param in schema._params.iteritems():
                self.check_schema(param, results, disabled_rules)

        if (isinstance(schema, jsonschema.Merge)):
            self.check_schema(schema.refschema, results, disabled_rules)

        if hasattr(schema, 'relations') and len(schema.relations) > 0:
            for relation in schema.relations.items():
                # relation[1] is type of jsonschema.Relation
                self.check_schema(relation[1], results, disabled_rules)

        return results

    def run(self, sdef):
        """
        Run validation against the servicedef.
        :param sdef: service definition to run against
        :type sdef: reschema.servicedef.ServiceDef

        :return: A list of Result objects
        """

        results = []

        print('Checking top-level schema correctness')
        sdef_disabled = get_disabled(sdef)
        results.extend(self._run_rules(Validator.SERVICEDEF_RULES, sdef,
                                       sdef_disabled))

        print('Checking types')
        for typedef in sdef.type_iter():
            if type(typedef) is not jsonschema.Ref:
                typedef_disabled = get_disabled(typedef).union(sdef_disabled)
                results.extend(self._run_rules(Validator.TYPE_RULES +
                                               Validator.SCHEMA_RULES,
                                               typedef, typedef_disabled))
                self.check_sub_schema(typedef, results, typedef_disabled)

        print('Checking resources')
        for resource in sdef.resource_iter():
            if type(resource) is not jsonschema.Ref:
                resource_disabled = get_disabled(resource).union(sdef_disabled)
                results.extend(self._run_rules(Validator.RESOURCE_RULES +
                                               Validator.SCHEMA_RULES,
                                               resource, resource_disabled))

                self.check_sub_schema(resource, results, resource_disabled)

            if self.VERBOSITY > 1:
                print('Checking links for \'{}\''.format(resource.name))

            for _, link in resource.links.items():
                if type(link) is not jsonschema.Ref:
                    link_disabled = get_disabled(link).union(resource_disabled)
                    results.extend(self._run_rules(Validator.LINK_RULES +
                                                   Validator.SCHEMA_RULES,
                                                   link, link_disabled))
                    self.check_sub_schema(link, results, link_disabled)
        return results

    def _run_rules(self, rules, obj, disabled_rules):
        """
        Runs a block of rules on a schema object

        :param rules: List of rules to run
        :param obj: reschema object to run the rules on

        :return: List of Result objects
        """

        results = []
        for rule in rules:
            try:
                rule(obj, disabled_rules)
                result = Result(rule.rule_id, obj.id, Result.PASSED)

            except ValidationFail as exc:
                result = Result(rule.rule_id, obj.id, Result.FAILED,
                                exc.message, exc)

            except RuleDisabled:
                result = Result(rule.rule_id, obj.id, Result.DISABLED)

            if (result.status != Result.PASSED) or (self.VERBOSITY > 1):
                print(str(result))

            results.append(result)

        return results


def get_disabled(obj):
    """
    Obtain the top-level disabled rules as well as on the object

    :param: schema object to get disabled rules from
    :type: reschema.jsonschema.Relation or
           reschema.jsonschema.Object or
           reschema.jsonschema.Link or
           reschema.servicedef.ServiceDef

    :returns: set of disabled rules of the object
    """
    return set(obj.tags.get('relint-disable', []))


def lint(sdef, filename):
    """
    Performs all checks on a loaded service definition.

    :param sdef:
    :type sdef: reschema.servicedef.ServiceDef

    :param filename:
    :type string: the absolute name of the processing yml file

    :returns: total number of failures
    """

    validator = Validator()
    results = validator.run(sdef)

    failures = len([r for r in results if r.status == Result.FAILED])

    # TODO: load cross-referenced schemas
    print('Checking referenced types and resources')
    errors = sdef.check_references()
    xref_failures = 0
    if errors:
        for error in errors:
            print('FAIL: [E0001] - \'{0}\' cannot be resolved'
                  .format(error.id))
        xref_failures += len(errors)

    errors = check_indentation(filename)
    indent_failures = len(errors)
    for row, col in errors:
        print ("{0}:{1}:{2} FAIL: [C0007] - indentation should be {3} spaces"
               .format(filename, row, col, INDENT_OFFSET))

    return xref_failures + failures + indent_failures


def first_char_pos(line):
    """
    return the indentation of first non-space character of the line
    If the first char is # or none such char exists, return False
    """
    striped_line = str.lstrip(line)
    if len(striped_line) == 0 or striped_line[0] == '#':
        return -1
    return len(line) - len(striped_line)


def check_indentation(filename):
    """
    check if each indentation is 4 spaces, otherwise generate issue C0007
    rule: when the first non-whitespace character is '#', ignore.
          otherwise, the indentation of the first valid char is deeper
          than the previous indentation, then the delta must be 4 spaces
    """
    try:
        with open(filename, 'r') as f:
            list = f.read().split('\n')
            ind = 0

            while ind < len(list) and first_char_pos(list[ind]) == -1:
                ind += 1
            pre_pos = first_char_pos(list[ind])
            curr_pos = 0
            errors = []
            while ind < len(list):
                curr_pos = first_char_pos(list[ind])
                ind += 1
                if curr_pos == -1:
                    continue
                if not (curr_pos == pre_pos or
                        curr_pos == pre_pos + INDENT_OFFSET or
                        (pre_pos > curr_pos and
                         (pre_pos - curr_pos) % INDENT_OFFSET == 0)):
                    errors.append((ind, curr_pos))
                pre_pos = curr_pos
            return errors
    except:
        print ("Function check_indentation failed, Exception: {0}"
               .format(traceback.format_exc()))
        return []


def check_valid_identifier(name, location):
    """
    Check that an identifier (link name, resource name, etc) is valid
    """
    if not re.match('^[a-z0-9_]*$', name):
        msg = "'{0}' contains invalid characters (only a-z, 0-9, _)"
        raise ValidationFail(msg.format(location), name)

    if not re.match('^[a-z]', name):
        msg = "'{0}' does not start with a letter"
        raise ValidationFail(msg.format(location), name)


def check_name_length(name, location):
    """
    Check that a name is at least 2 characters long and contain a vowel
    """
    if len(name) < 2:
        raise ValidationFail(
            "'{0}' should be at least 2 characters long".format(location.id),
            name)


def check_valid_description(text, location, obj=None, required=True):
    """
    Checks the text of a description
    """
    if not text and required:
        raise ValidationFail("'{0}' has no description field".format(location),
                             obj)

    if not text[0].isupper():
        msg = "'{0}' description does not begin with uppercase character"
        raise ValidationFail(msg.format(location), text)


@Validator.servicedef('W0001')
def schema_provider_valid(sdef):
    if sdef.provider != 'riverbed':
        raise ValidationFail("'provider' field must be 'riverbed'",
                             sdef.provider)


@Validator.servicedef('W0002')
def schema_id_valid(sdef):
    id_re = Validator.RVBD_BASE_URL + '/([a-z0-9_.]+)/([0-9](.[0-9]){1,2})$'
    match = re.match(id_re, sdef.id)
    if not match:
        msg = "'id' must be '{0}/<name>/<version>'"
        raise ValidationFail(msg.format(Validator.RVBD_BASE_URL), sdef.id)
    name = match.group(1)
    if name != sdef.name:
        msg = "schema id does not match the schema name: '{0}'"
        raise ValidationFail(msg.format(sdef.name), sdef.id)
    version = match.group(2)
    if version != sdef.version:
        msg = "schema id does not match the schema version: '{0}'"
        raise ValidationFail(msg.format(sdef.version), sdef.id)


@Validator.servicedef('W0003')
def schema_field_valid(sdef):
    schema_re = Validator.RVBD_BASE_URL + '/service_def/([0-9](.[0-9]){1,2})$'
    match = re.match(schema_re, sdef.schema)
    if not match:
        msg = "'$schema' field must be '{0}/service_def/<version>'"
        raise ValidationFail(msg.format(Validator.RVBD_BASE_URL), sdef.schema)


@Validator.servicedef('W0004')
def schema_has_title(sdef):
    if not sdef.title:
        raise ValidationFail("the schema must have a title", sdef)


@Validator.schema('W0005')
def schema_has_additional_properties(schema):
    if isinstance(schema, jsonschema.Object):
        if 'additionalProperties' not in schema.input:
            raise ValidationFail("additionalProperties missing in '{0}'"
                                 .format(schema.id), schema.name)


@Validator.schema('W0006')
def required_property_with_default_value(schema):
    if hasattr(schema, 'required') and schema.required:
        for required_prop in schema.required:
            if required_prop in schema.properties:
                default = getattr(schema.properties[required_prop], 'default',
                                  None)
                if default is not None:
                    raise ValidationFail(
                        "A required property '{0}' should not have a default "
                        "value in '{1}'".format(
                            required_prop, schema.fullname()
                        ),
                        default)


@Validator.schema('C0001')
def schema_has_valid_name(schema):
    # introducing the below logic because schemas underneath any/one/allof
    # are with names as any/one/allOf[*], where * is a non-negative int
    # for those schema object, we do not want to generate C0001 failures
    if hasattr(schema, 'parent') and hasattr(schema.parent, 'input'):
        keys = schema.parent.input.keys()
        if set(['anyOf', 'oneOf', 'allOf']) & set(keys):
            return

    check_valid_identifier(schema.name, schema.id)


@Validator.typedef('C0002')
def type_does_not_have_prefix_or_suffix(typedef):
    # Some people suffix or prefix the name with type, not needed with the
    # 2.1 $ref: #/types syntax
    type_lc = typedef.name.lower()
    if type_lc.startswith('type'):
        raise ValidationFail(
            "'{0}' should not start with 'type'".format(typedef.id),
            typedef.name)
    elif type_lc.endswith('type'):
        raise ValidationFail(
            "'{0}' should not end with 'type'".format(typedef.id),
            typedef.name)


@Validator.resource('C0003')
def resource_does_not_have_prefix_or_suffix(resource):
    res_lc = resource.name.lower()
    if res_lc.startswith('resource'):
        raise ValidationFail(
            "'{0}' should not start with 'resource'".format(resource.id),
            resource.name)
    elif res_lc.endswith('resource'):
        raise ValidationFail(
            "'{0}' should not end with 'resource'".format(resource.id),
            resource.name)


@Validator.link('C0004')
def link_does_not_have_prefix_or_suffix(link):
    link_lc = link.name.lower()
    if link_lc.startswith('link'):
        raise ValidationFail(
            "'{0}' should not start with 'link'".format(link.id),
            link.name)
    elif link_lc.endswith('link'):
        raise ValidationFail(
            "'{0}' should not end with 'link'".format(link.id),
            link.name)


@Validator.schema('C0005')
def schema_name_check_length(schema):
    check_name_length(schema.name, schema)


@Validator.servicedef('C0006')
def sd_has_valid_description(sdef):
    check_valid_description(sdef.description, 'ServiceDef', sdef,
                            required=True)


@Validator.link('C0100')
def link_std_has_no_description(link):
    if link.name in Validator.STD_LINKS and link.description:
        raise ValidationFail(
            "'{0}' link must have no description".format(link.name),
            link.description)


@Validator.link('C0101')
def link_non_std_has_description(link):
    if link.name not in Validator.STD_LINKS:
        check_valid_description(link.description, link.id, link, required=True)


@Validator.link('W0100')
def link_get_has_no_request_body(link):
    if 'get' == link.name and 'null' != link.request.typestr:
        obj = get_first_marked_object(link._request, link.name)
        raise ValidationFail("'get' link cannot have a request body", obj)


@Validator.link('W0101')
def link_get_response_is_the_resource(link):
    if 'get' == link.name and link.schema.id != link.response.id:
        obj = get_first_marked_object(link._response, link.name)
        raise ValidationFail(
            "'get' link response must be '{0}' ".format(link.schema.id),
            obj)


@Validator.link('W0102')
def link_set_request_is_the_resource(link):
    if 'set' == link.name and link.schema.id != link.request.id:
        obj = get_first_marked_object(link._request, link.name)
        raise ValidationFail(
            "'set' link request must be '{0}' ".format(link.schema.id),
            obj)


@Validator.link('W0103')
def link_set_response_is_null_or_resource(link):
    if ('set' == link.name and
            'null' != link.response.typestr and
            link.schema.id != link.response.id):
        obj = get_first_marked_object(link._response, link.name)
        msg = "'set' link response must be empty or '{0}'"
        raise ValidationFail(msg.format(link.schema.id), obj)


@Validator.link('W0104')
def link_delete_has_no_request_body(link):
    if 'delete' == link.name and 'null' != link.request.typestr:
        obj = get_first_marked_object(link._request, link.name)
        raise ValidationFail("'delete' link cannot have a request body", obj)


@Validator.link('W0105')
def link_delete_has_no_response_body(link):
    if 'delete' == link.name and 'null' != link.response.typestr:
        obj = get_first_marked_object(link._response, link.name)
        raise ValidationFail("'delete' link cannot have a response body", obj)


@Validator.link('W0106')
def link_create_has_request_body(link):
    if 'create' == link.name and 'null' == link.request.typestr:
        raise ValidationFail("'create' link must have a request body",
                             link.name)


@Validator.link('W0107')
def link_create_request_is_not_resource(link):
    if 'create' == link.name and link.schema.id == link.request.id:
        raise ValidationFail(
            "'create' request must not be '{0}'".format(link.schema.id),
            link._request)


@Validator.link('W0108')
def link_create_response_is_not_resource(link):
    if 'create' == link.name and link.schema.id == link.response.id:
        raise ValidationFail(
            "'create' response must not be '{0}'".format(link.schema.id),
            link._response)


@Validator.link('W0112')
def link_path_does_not_end_in_slash(link):
    if link.path:
        pathstr = str(link.path)
        if pathstr.endswith('/') and pathstr != '/':
            raise ValidationFail("A link path must not end with /", link.path)


@Validator.link('E0100')
def link_get_method_is_get(link):
    if 'get' == link.name and 'GET' != link.method.upper():
        raise ValidationFail("'get' link must use http method GET",
                             link.method)


@Validator.link('E0101')
def link_set_method_is_put(link):
    if 'set' == link.name and 'PUT' != link.method.upper():
        raise ValidationFail("'set' link must use http method PUT",
                             link.method)


@Validator.link('E0102')
def link_create_method_is_post(link):
    if 'create' == link.name and 'POST' != link.method.upper():
        raise ValidationFail("'create' link must use http method POST",
                             link.method)


@Validator.link('E0103')
def link_delete_method_is_delete(link):
    if 'delete' == link.name and 'DELETE' != link.method.upper():
        raise ValidationFail("'delete' link must use http method DELETE",
                             link.method)


@Validator.resource('C0300')
def resource_has_valid_description(resource):
    check_valid_description(resource.description, resource.id, resource,
                            required=True)


@Validator.resource('C0301')
def resource_type_is_object(resource):
    # NOTE: relint does not load referenced files yet.
    if resource.is_ref():
        print("Referenced schema not validated: '{0}'".format(
              resource.refschema.fullid()))
        return
    if 'object' != resource.typestr:
        raise ValidationFail(
            "resource '{0}' should be an object".format(resource.name),
            resource)


@Validator.typedef('C0200')
def type_has_valid_description(typedef):
    check_valid_description(typedef.description, typedef.id, typedef,
                            required=True)


@Validator.schema('E0002')
def required_is_valid(schema):
    if (hasattr(schema, 'required') and hasattr(schema, 'properties')
            and hasattr(schema, 'additional_properties')):
        if (schema.additional_properties is False and
                schema.required is not None):
            for k in schema.required:
                if k not in schema.properties:
                    raise ValidationFail("Required field '{0}' is not"
                                         " defined in '{1}'"
                                         .format(k, schema.fullname()), k)


@Validator.schema('E0003')
def relation_is_valid(schema):
    if isinstance(schema, jsonschema.Relation):
        try:
            schema.resource
        except jsonschema.InvalidReference:
            raise ValidationFail("Invalid relation '{0}': '{1}' not found".
                                 format(schema.fullname(),
                                        schema._resource_id), schema.name)


@Validator.link('E0105')
def link_uritemplate_param_declared(link):
    params = uritemplate.variables(link.path.template)
    for param in params:
        if link.schema.typestr != 'object':
            msg = ("The resource '{0}' is not an object, "
                   "so the parameter '{1}' in the uritemplate "
                   "'{2}' is not declared in link '{3}'")
            raise ValidationFail(
                msg.format(link.schema.fullname(), param, link.path.template,
                           link.name),
                link)
        elif param not in link.schema.properties:
            msg = ("The parameter '{0}' in the uritemplate '{1}'"
                   " in link '{2}' is not declared in properties"
                   " in resouce '{3}'")
            raise ValidationFail(
                msg.format(param, link.path.template, link.name,
                           link.schema.fullname()),
                link)


@Validator.resource('C0303')
def self_link_is_first(resource):
    if 'self' not in resource.links:
        return

    self_mark = resource.links['self'].name.start_mark
    for link in resource.links.values():
        mark = link.name.start_mark
        if link.name == "self":
            continue
        if (mark.line, mark.column) < (self_mark.line, self_mark.column):
            msg = "'self' link should be the first in '{0}'"
            raise ValidationFail(
                msg.format(resource.fullname()),
                resource.links['self'].name
            )



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import os
import unittest

from reschema.exceptions import ReschemaException
from reschema import ServiceDef, ServiceDefManager

"""
Exmaple script to demonstrate how marks can print out location
of parsing problems from source data.
"""

yaml_snippet_bad = """\
$schema: "service_def/2.0"
name: "Bookstore"
version: "1.0"
title: "REST API for a bookstore"
servicePath: "/api/bookstore/1.0"
defaultAuthorization: "required"
resources:
   info:
      type: object
      invalid_type: foo
      properties:
         owner: { type: string }
         email: { type: string }

      links:
         self: { path: "$/info" }
         books: { target: books }
         authors: { target: authors }
         get:
            method: GET
            response: { $ref: info }
         set:
            method: PUT
            request: { $ref: info }
            response: { $ref: info }
"""

json_snippet_bad = """\
{
  "$schema": "service_def/2.0",
  "name": "Bookstore",
  "version": "1.0",
  "title": "REST API for a bookstore",
  "servicePath": "/api/bookstore/1.0",
  "defaultAuthorization": "required",
  "resources": {
    "info": {
      "type": "object",
      "invalid_type": "foo",
      "properties": {
        "owner": {
          "type": "string"
        },
        "email": {
          "type": "string"
        }
      },
      "links": {
        "self": {
          "path": "$/info"
        },
        "books": {
          "target": "books"
        },
        "authors": {
          "target": "authors"
        },
        "get": {
          "method": "GET",
          "response": {
            "$ref": "info"
          }
        },
        "set": {
          "method": "PUT",
          "request": {
            "$ref": "info"
          },
          "response": {
            "$ref": "info"
          }
        }
      }
    }
  }
}
"""


class TestMarkedLoad(unittest.TestCase):

    def test_bad_yaml(self):
        with self.assertRaises(ReschemaException):
            r = ServiceDef()
            r.parse_text(yaml_snippet_bad, format='yaml')

    def test_bad_json(self):
        with self.assertRaises(ReschemaException):
            r = ServiceDef()
            r.parse_text(json_snippet_bad)



#!/usr/bin/env python

# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import sys
import json

from reschema.yaml_loader import marked_load

"""
Takes YAML input file and prints out a JSON representation to stdout.
"""


def convert(stream):
    yml = marked_load(stream)
    s = json.dumps(yml, indent=2).split("\n")
    result = []
    for line in s:
        result.append(line.rstrip())
    return '\n'.join(result)


if __name__ == '__main__':
    yaml_file = sys.argv[1]
    with open(yaml_file, 'r') as f:
        print convert(f)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import mock
import pytest

from reschema import jsonschema


ANY_NAME = 'foo'
ANY_PARENT_NAME = 'bar'
ANY_ID = '#/resources/foo'
ANY_SERVICEDEF_ID = 'http://support.riverbed.com/apis/mock/1.0'

ANY_INTERMEDIARY = 'inter'
ANY_INPUT = {'blah': 'whatever'}

MOCK_SERVICEDEF = mock.Mock()
MOCK_SERVICEDEF.id = ANY_SERVICEDEF_ID
MOCK_SERVICEDEF.fullid.return_value = ANY_SERVICEDEF_ID

MOCK_PARENT = mock.Mock()
MOCK_PARENT.name = ANY_PARENT_NAME
MOCK_PARENT.fullname.return_value = ANY_PARENT_NAME
MOCK_PARENT.servicedef = MOCK_SERVICEDEF

ANY_FULL_ID = '/'.join((MOCK_SERVICEDEF.id, ANY_ID))


@pytest.fixture
def e_parent_only():
    e = jsonschema.Entity(id=ANY_ID,
                          name=ANY_NAME,
                          parent=MOCK_PARENT)
    return e


def test_init_parent_only(e_parent_only):
    assert e_parent_only.id == ANY_ID
    assert e_parent_only.name == ANY_NAME
    assert e_parent_only.parent is MOCK_PARENT
    assert e_parent_only.servicedef is MOCK_SERVICEDEF


def test_repr(e_parent_only):
    assert repr(e_parent_only) == "<servicedef.%s '%s'>" % (
        jsonschema.Entity.__name__, e_parent_only.fullid())


def test_absolute_id(e_parent_only):
    assert e_parent_only.fullid() == MOCK_SERVICEDEF.id + ANY_ID


def test_relative_id(e_parent_only):
    assert e_parent_only.fullid(True) == ANY_ID


# fullname() was originally defined separately in many places, hence the
# various specific unit tests.  The behaviors should continue with the
# base class Entity

def test_name_schema_no_parent():
    e = jsonschema.Entity(id=ANY_ID, name=ANY_NAME,
                          servicedef=MOCK_SERVICEDEF)
    assert e.fullname() == e.name


def test_name_no_parent_no_name():
    unnamed_parent = mock.Mock()
    unnamed_parent.servicedef = MOCK_SERVICEDEF
    unnamed_parent.fullname.return_value = None

    e = jsonschema.Entity(id=ANY_ID, name=None, parent=unnamed_parent)
    assert e.fullname() is None


def test_name_schema_parent_unnamed():
    unnamed_parent = mock.Mock()
    unnamed_parent.servicedef = MOCK_SERVICEDEF
    unnamed_parent.fullname.return_value = None

    e = jsonschema.Entity(id=ANY_ID, name=ANY_NAME, parent=unnamed_parent)
    assert e.fullname() == e.name


def test_name_schema_parent_named(e_parent_only):
    e = e_parent_only
    assert e.fullname() == '%s.%s' % (e.parent.fullname(), e.name)


def test_name_schema_parent_array():
    array = mock.Mock(spec=jsonschema.Array, instance=True)
    array.fullname.return_value = ANY_PARENT_NAME
    array.servicedef = MOCK_SERVICEDEF
    e = jsonschema.Entity(id=ANY_ID, name=ANY_NAME, parent=array)
    assert e.fullname() == '%s[%s]' % (e.parent.fullname(), e.name)


def test_name_intermediary():
    e = jsonschema.Entity(id=ANY_ID, name=ANY_NAME, parent=MOCK_PARENT,
                          intermediary=ANY_INTERMEDIARY)
    assert e.fullname() == '%s.%s.%s' % (ANY_PARENT_NAME,
                                         ANY_INTERMEDIARY,
                                         ANY_NAME)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from __future__ import print_function

import os
import logging
import unittest

import reschema
from reschema.lint import Validator, Result

logger = logging.getLogger(__name__)

TEST_PATH = os.path.abspath(os.path.dirname(__file__))
SERVICE_DEF_TEST = os.path.join(TEST_PATH, 'service_test.yaml')
SERVICE_DEF_TEST_REF = os.path.join(TEST_PATH, 'service_test_ref.yaml')

SERVICE_DEF_TEMPLATE = '''
$schema: http://support.riverbed.com/apis/service_def/2.2
id: http://support.riverbed.com/apis/test/1.0
provider: riverbed
name: test
version: '1.0'
tags: {}
'''


def create_servicedef(fragment):
    """
    Parse a schema fragment and create a ServiceDef instance that can be
    passed to the Validator.

    The fragment should be from an outermost section of a servicedef file, eg
    "types:
       test:
         type: integer"

    The fragment will be merged into a template servicedef instance; any keys
    in the fragment override/are added to the template.

    :param fragment: Schema fragment to merge in
    :return: TestValidator
    """
    sdef = reschema.ServiceDef.create_from_text(SERVICE_DEF_TEMPLATE +
                                                fragment,
                                                format='yaml')

    return sdef


class TestLintBase(unittest.TestCase):
    """
    Helper class to run and validate relint calls
    """

    def check_result(self, rule_id, obj_id, result, fragment):
        """
        Verifies that relint declares  on the given fragment

        :param rule_id: Rule to verify passed
        :param obj_id: Object to verify rule passed on
        :param result: Test result expected
        :param fragment: fragment to check
        """

        sdef = create_servicedef(fragment)
        validator = Validator()
        results = validator.run(sdef)

        found_result = None
        for r in results:
            if (r.rule_id == rule_id) and (r.obj_id == obj_id):
                found_result = r
                break

        self.assertTrue(found_result,
                        'No result for {} on {} found'.format(rule_id, obj_id))
        self.assertEqual(found_result.status, result,
                         'Unexpected {}'.format(found_result))


class TestRelintDisable(TestLintBase):
    """
    Collection of tests that verify the relint-disable handling
    """

    def test_servicedef_disable(self):
        """
        Verifies top-level servicedef checks honor disable
        """
        self.check_result('W0002',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.DISABLED,
                          'tags:\n'
                          '   relint-disable: [ W0002 ] ')

    def test_resource_disable(self):
        """
        Verifies resource-level checks can be disabled
        """

        # Tag disable at the top level
        self.check_result('C0003', '#/resources/foo_resource', Result.DISABLED,
                          'tags:\n'
                          '   relint-disable: [ C0003 ]\n'
                          '\n'
                          'resources:\n'
                          '  foo_resource:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }')

        # Tag disable on the resource
        self.check_result('C0003', '#/resources/foo_resource', Result.DISABLED,
                          'resources:\n'
                          '  foo_resource:\n'
                          '    tags:\n'
                          '       relint-disable: [ C0003 ]\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }')

        # Tag disable on the sub-schema of resource
        self.check_result('C0001', '#/resources/foo_resource/properties/name',
                          Result.DISABLED,
                          'resources:\n'
                          '  foo_resource:\n'
                          '    tags:\n'
                          '       relint-disable: [ C0001 ]\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      name:\n'
                          '        type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }')

    def test_type_disable(self):
        """
        Verifies type-level checks can be disabled
        """

        # Tag disable at the top level
        self.check_result('C0002', '#/types/foo_type', Result.DISABLED,
                          'tags:\n'
                          '   relint-disable: [ C0002 ]\n'
                          '\n'
                          'types:\n'
                          '  foo_type:\n'
                          '    type: string')

        # Tag disable on the type
        self.check_result('C0002', '#/types/foo_type', Result.DISABLED,
                          'types:\n'
                          '  foo_type:\n'
                          '    tags:\n'
                          '       relint-disable: [ C0002 ]\n')

        # test disable on the sub-schema of type
        self.check_result('C0002', '#/types/foo_type/properties/name',
                          Result.DISABLED,
                          'types:\n'
                          '  foo_type:\n'
                          '    tags:\n'
                          '       relint-disable: [ C0002 ]\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      name: {type: string}\n')

    def test_link_disable(self):
        """
        Verifies link-level checks can be disabled
        """

        # Tag disable at the top level
        self.check_result('C0004', '#/resources/foo/links/foo_link',
                          Result.DISABLED,
                          'tags:\n'
                          '   relint-disable: [ C0004 ]\n'
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '      foo_link:\n'
                          '        path: /foo/nope\n'
                          '        method: GET')

        # Tag disable on the resource
        self.check_result('C0004', '#/resources/foo/links/foo_link',
                          Result.DISABLED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    tags:\n'
                          '       relint-disable: [ C0004 ]\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '      foo_link:\n'
                          '        path: /foo/nope\n'
                          '        method: GET')

        # Tag disable on the link
        self.check_result('C0004', '#/resources/foo/links/foo_link',
                          Result.DISABLED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '      foo_link:\n'
                          '        path: /foo/nope\n'
                          '        method: GET\n'
                          '        tags:\n'
                          '           relint-disable: [ C0004 ]')

        # test disable on the subschema of link
        self.check_result('C0005',
                          '#/resources/foo/links/link1/request/properties/p1',
                          Result.DISABLED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '      link1:\n'
                          '        path: /foo/nope\n'
                          '        method: GET\n'
                          '        tags:\n'
                          '           relint-disable: [ C0005 ]\n'
                          '        request:\n'
                          '          type: object\n'
                          '          properties:\n'
                          '            p1: { type: number }')


class TestRelint(TestLintBase):
    def test_rule_W0001(self):
        ''' The ``provider`` field must be set to ``riverbed`` '''

        self.check_result('W0001',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.PASSED,
                          'provider: riverbed\n')

        self.check_result('W0001',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'provider: wrong_provider\n')

    def test_rule_W0002(self):
        ''' The ``id`` field must be
        ``http://support.riverbed.com/apis/{name}/{version}`` '''

        self.check_result('W0002',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.PASSED,
                          'id: http://support.riverbed.com/apis/test/1.0\n'
                          'name: test')

        self.check_result('W0002',
                          'http://support.riverbed.com/apis/wrongid/1.0',
                          Result.FAILED,
                          'id: http://support.riverbed.com/apis/wrongid/1.0\n'
                          'name: test')

        self.check_result('W0002',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'id: http://support.riverbed.com/apis/test/1.0\n'
                          'name: wrongname')

    def test_rule_W0003(self):
        ''' the schema field must be set to
            http://support.riverbed.com/apis/service_def/{version}
        '''
        self.check_result('W0003',
                          'http://support.riverbed.com/apis/service_def/2.2',
                          Result.PASSED,
                          "$schema: "
                          "http://support.riverbed.com/apis/service_def/2.2\n"
                          'id: '
                          'http://support.riverbed.com/apis/service_def/2.2\n'
                          'name: test\n')

    def test_rule_W0004(self):
        ''' the schema must have a title '''

        self.check_result('W0004',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.PASSED,
                          'title: test_relint\n')

        self.check_result('W0004',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'title: \n')

        self.check_result('W0004',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'id: http://support.riverbed.com/apis/test/1.0\n'
                          'name: test\n')

    def test_rule_W0005(self):
        '''additionalProperties required for Object schema'''
        self.check_result('W0005', '#/resources/info',
                          Result.PASSED,
                          'resources:\n'
                          '  info:\n'
                          '    type: object\n'
                          '    additionalProperties : False\n')

        self.check_result('W0005', '#/resources/info',
                          Result.FAILED,
                          'resources:\n'
                          '  info:\n'
                          '    type: object\n')

    def test_rule_C0001(self):
        '''name should start with a letter and contains only
           lowercase, numbers and _
        '''
        self.check_result('C0001', '#/types/foo1_', Result.PASSED,
                          'types:\n'
                          '  foo1_: { type: string }')

        self.check_result('C0001', '#/types/1foo', Result.FAILED,
                          'types:\n'
                          '  1foo: { type: string }')

        self.check_result('C0001', '#/types/Foo', Result.FAILED,
                          'types:\n'
                          '  Foo: { type: string }')

        self.check_result('C0001', '#/types/foo!', Result.FAILED,
                          'types:\n'
                          '  foo!: { type: string }')

    def test_rule_C0002(self):
        """ Type should not end in _type """

        self.check_result('C0002', '#/types/foo', Result.PASSED,
                          'types:\n'
                          '  foo: { type: string }')

        self.check_result('C0002', '#/types/foo_type', Result.FAILED,
                          'types:\n'
                          '  foo_type: { type: string }')

    def test_rule_C0003(self):
        """ Resource should not end in _resource """

        self.check_result('C0003', '#/resources/foo', Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }')

        self.check_result('C0003', '#/resources/foo_resource', Result.FAILED,
                          'resources:\n'
                          '  foo_resource:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }')

    def test_rule_C0004(self):
        ''' The name of a link should not start or end with ``link``'''

        self.check_result('C0004', '#/resources/foo/links/bar',
                          Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '      bar:\n'
                          '        path: /foo/yep\n'
                          '        method: GET')

        self.check_result('C0004', '#/resources/foo/links/foo_link',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '      foo_link:\n'
                          '        path: /foo/nope\n'
                          '        method: GET')

    def test_rule_C0005(self):
        ''' A resource, type, link, or relation name must be
            at least 2 characters long
        '''

        self.check_result('C0005', '#/types/a_good_type_name',
                          Result.PASSED,
                          'types:\n'
                          '  a_good_type_name:\n'
                          '    type: object\n')

        self.check_result('C0005', '#/types/t',
                          Result.FAILED,
                          'types:\n'
                          '  t: \n'
                          '    type: object\n')

        self.check_result('C0005', '#/resources/a_good_resource_name',
                          Result.PASSED,
                          'resources:\n'
                          '  a_good_resource_name:\n'
                          '    type: object\n')

        self.check_result('C0005', '#/resources/r',
                          Result.FAILED,
                          'resources:\n'
                          '  r:\n'
                          '    type: object\n')

        self.check_result('C0005', '#/resources/res/links/a_good_link_name',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self: { path: /res }\n'
                          '      a_good_link_name:\n'
                          '        method: GET\n')

        self.check_result('C0005', '#/resources/res/links/l',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self: { path: /res }\n'
                          '      l:\n'
                          '        method: GET\n')

    def test_rule_C0006(self):
        ''' The service definition must have a valid description field,
            starting with a capital letter
        '''

        self.check_result('C0006',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.PASSED,
                          'description: Test description')

        self.check_result('C0006',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'description: test description')

        self.check_result('C0006',
                          'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'description: ')

        self.check_result('C0006', 'http://support.riverbed.com/apis/test/1.0',
                          Result.FAILED,
                          'id: http://support.riverbed.com/apis/test/1.0\n'
                          'name: test\n')

    def test_rule_E0002(self):
        ''' Required fields should exist in properties if
            additionalProperties is False
        '''
        self.check_result('E0002', '#/types/foo', Result.PASSED,
                          'types:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      name: { type: string }\n'
                          '    required: [ name ]\n'
                          '    additionalProperties: false')

        self.check_result('E0002', '#/types/foo', Result.FAILED,
                          'types:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      name: { type: string }\n'
                          '    required: [ nonesuch ]\n'
                          '    additionalProperties: false')

        self.check_result('E0002', '#/types/foo', Result.PASSED,
                          'types:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      name: { type: string }\n'
                          '    required: [ nonesuch ]\n'
                          '    additionalProperties: true')

        self.check_result('E0002', '#/resources/foo', Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '    required: [id]\n'
                          '    additionalProperties: false')

        self.check_result('E0002', '#/resources/foo', Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '    required: [id, non_existing_prop]\n'
                          '    additionalProperties: False')

        # check the recursive
        self.check_result('E0002', '#/resources/foo', Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '      outer:\n'
                          '        type: object\n'
                          '        properties:\n'
                          '          inner: { type: number } \n'
                          '        required: [inner]\n'
                          '        additionalProperties: False\n'
                          '    links:\n'
                          '      self: { path: /foo }\n')

        self.check_result('E0002', '#/resources/foo/properties/outer',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '      outer:\n'
                          '        type: object\n'
                          '        properties:\n'
                          '          inner: { type: number } \n'
                          '        required: [inner, non_present]\n'
                          '        additionalProperties: False\n'
                          '    links:\n'
                          '      self: { path: /foo }\n')

        self.check_result('E0002', '#/resources/foo/properties/outer/items',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '      outer:\n'
                          '        type: array\n'
                          '        items:\n'
                          '           type: object\n'
                          '           properties:\n'
                          '              inner: { type: number } \n'
                          '           required: [inner, non_present]\n'
                          '           additionalProperties: False\n'
                          '    links:\n'
                          '      self: { path: /foo }\n')

    def test_rule_E0003(self):
        '''relations should be valid. The specified resource must be found'''

        self.check_result('E0003', '#/resources/info/relations/foo',
                          Result.PASSED,
                          'resources:\n'
                          '  info:\n'
                          '    relations:\n'
                          '      foo:\n'
                          '        resource: \'#/resources/info\'\n')

        self.check_result('E0003', '#/resources/info/relations/foo',
                          Result.FAILED,
                          'resources:\n'
                          '  info:\n'
                          '    relations:\n'
                          '      foo:\n'
                          '        resource: \'#/resources/foo\'\n')

        # when relations is nested inside of each element of an array
        self.check_result('E0003', '#/resources/info/items/relations/foo',
                          Result.PASSED,
                          'resources:\n'
                          '  info:\n'
                          '    type: array\n'
                          '    items:\n'
                          '      relations:\n'
                          '        foo:\n'
                          '          resource: \'#/resources/info\'\n')

        self.check_result('E0003', '#/resources/info/items/relations/foo',
                          Result.FAILED,
                          'resources:\n'
                          '  info:\n'
                          '    type: array\n'
                          '    items:\n'
                          '      relations:\n'
                          '        foo:\n'
                          '          resource: \'#/resources/foo\'\n')

    def test_rule_C0100(self):
        '''  Standard links must not have a description field. '''

        self.check_result('C0100', '#/resources/res/links/self',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n')

        self.check_result('C0100', '#/resources/res/links/self',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '        description: needless description\n')

    def test_rule_C0101(self):
        ''' A non-standard link must have a valid description field. '''

        self.check_result('C0101', '#/resources/res/links/buy',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      buy:\n'
                          '        method: GET\n'
                          '        path: "/path"\n'
                          '        description: non capital description\n')

        self.check_result('C0101', '#/resources/res/links/buy',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      buy:\n'
                          '        method: GET\n'
                          '        path: "/path"\n'
                          '        description: buy it\n')

        self.check_result('C0101', '#/resources/res/links/buy',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      buy:\n'
                          '        method: GET\n'
                          '        path: "/path"\n'
                          '        description: Capital description\n')

    def test_rule_W0100(self):
        '''  A ``get`` link cannot have a request body '''

        self.check_result('W0100', '#/resources/res/links/get',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      get:\n'
                          '        method: GET\n'
                          '        request: { type: string }\n')

        self.check_result('W0100', '#/resources/res/links/get',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      get:\n'
                          '        method: GET\n')

    def test_rule_W0101(self):
        '''  A ``get`` link cannot have a request body '''

        self.check_result('W0101', '#/resources/res/links/get',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      get:\n'
                          '        method: GET\n'
                          '        response: { type: string }\n')

        self.check_result('W0101', '#/resources/res/links/get',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      get:\n'
                          '        method: GET\n'
                          '        response: { $ref: "#/resources/res" }\n')

    def test_rule_W0102(self):
        '''A ``set`` link request must be the representation
           of the resource it belongs to
        '''

        self.check_result('W0102', '#/resources/res/links/set',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: PUT\n'
                          '        request: { $ref: "#/resources/res" }\n')

        self.check_result('W0102', '#/resources/res/links/set',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: PUT\n'
                          '        request: { type: string }\n')

    def test_rule_W0103(self):
        '''A ``set`` link response must be null or the representation
          of the resource it belongs to
        '''

        self.check_result('W0103', '#/resources/res/links/set',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: PUT\n')

        self.check_result('W0103', '#/resources/res/links/set',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: PUT\n'
                          '        response: { type: string }\n')

        self.check_result('W0103', '#/resources/res/links/set',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: PUT\n'
                          '        response: { $ref: "#/resources/res" }\n')

    def test_rule_W0104(self):
        ''' A ``delete`` link cannot have a request body '''

        self.check_result('W0104', '#/resources/res/links/delete',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      delete:\n'
                          '        method: DELETE\n'
                          '        request: { type: string }\n')

        self.check_result('W0104', '#/resources/res/links/delete',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      delete:\n'
                          '        method: DELETE\n')

    def test_rule_W0105(self):
        ''' A ``delete`` link cannot have a response body '''

        self.check_result('W0105', '#/resources/res/links/delete',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      delete:\n'
                          '        method: DELETE\n'
                          '        response: { type: string }\n')

        self.check_result('W0105', '#/resources/res/links/delete',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      delete:\n'
                          '        method: DELETE\n')

    def test_rule_W0106(self):
        '''A ``create`` link must have a request body '''

        self.check_result('W0106', '#/resources/res/links/create',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n')

        self.check_result('W0106', '#/resources/res/links/create',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n'
                          '        request: { type: string }\n')

    def test_rule_W0107(self):
        ''' A ``create`` link request must not be the same
            as the resource it belongs to
        '''

        self.check_result('W0107', '#/resources/res/links/create',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n'
                          '        request: { $ref: "#/resources/res"}\n')

        self.check_result('W0107', '#/resources/res/links/create',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n'
                          '        request: { type: string }\n')

    def test_rule_W0108(self):
        '''A ``create`` link response must not be the same as
           the resource it belongs to
        '''

        self.check_result('W0108', '#/resources/res/links/create',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n'
                          '        response: { $ref: "#/resources/res"}\n')

        self.check_result('W0108', '#/resources/res/links/create',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n'
                          '        response: { type: string }\n')

    def test_rule_W0112(self):
        ''' A link should not end with / '''

        self.check_result('W0112', '#/resources/foo/links/self',
                          Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/foo"\n'
                          '      bar:\n'
                          '        method: POST\n'
                          '        path: /foo/bar\n')

        # Self link
        self.check_result('W0112', '#/resources/foo/links/self',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/foo/"\n'
                          '      bar:\n'
                          '        method: POST\n'
                          '        path: /foo/bar\n')

        # Bar link with inherited path
        self.check_result('W0112', '#/resources/foo/links/bar',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/foo/"\n'
                          '      bar:\n'
                          '        method: POST\n')

        # Bar link with explicit path
        self.check_result('W0112', '#/resources/foo/links/bar',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/foo"\n'
                          '      bar:\n'
                          '        method: POST\n'
                          '        path: /foo/bar/\n')

        # Special check for root
        self.check_result('W0112', '#/resources/foo/links/self',
                          Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: string\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/"\n')

    def test_rule_E0100(self):
        ''' A ``get`` link must use http method GET '''

        self.check_result('E0100', '#/resources/res/links/get',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      get:\n'
                          '        method: GET\n')

        self.check_result('E0100', '#/resources/res/links/get',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      get:\n'
                          '        method: PUT\n')

    def test_rule_E0101(self):
        '''  A ``set`` link must use http method PUT '''

        self.check_result('E0101', '#/resources/res/links/set',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: PUT\n')

        self.check_result('E0101', '#/resources/res/links/set',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      set:\n'
                          '        method: GET\n')

    def test_rule_E0102(self):
        '''  A ``create`` link must use http method POST '''

        self.check_result('E0102', '#/resources/res/links/create',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: POST\n')

        self.check_result('E0102', '#/resources/res/links/create',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      create:\n'
                          '        method: GET\n')

    def test_rule_E0103(self):
        '''  A ``delete`` link must use http method DELETE '''

        self.check_result('E0103', '#/resources/res/links/delete',
                          Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      delete:\n'
                          '        method: DELETE\n')

        self.check_result('E0103', '#/resources/res/links/delete',
                          Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self:\n'
                          '        path: "/path"\n'
                          '      delete:\n'
                          '        method: GET\n')

    def test_rule_E0105(self):
        '''A parameter in URI template must be declared in schema properties'''

        self.check_result('E0105', '#/resources/foo/links/self', Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '    links:\n'
                          '      self: { path: "/foos/{id}" }\n')
        self.check_result('E0105', '#/resources/foo/links/self', Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '    links:\n'
                          '      self: { path: "/foos/{non_present}" }\n')
        self.check_result('E0105', '#/resources/foo/links/self', Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: "null"\n'
                          '    links:\n'
                          '      self: { path: "/foos/{non_present}" }\n')

    def test_rule_C0303(self):
        ''' Self link should be the first link '''

        self.check_result('C0303', '#/resources/r1', Result.PASSED,
                          'resources:\n'
                          '  r1:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self: { path: "/r1" }\n')

        self.check_result('C0303', '#/resources/r4', Result.PASSED,
                          'resources:\n'
                          '  r4:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      r4_l1: { method: GET, path: "/r4" }\n')

        self.check_result('C0303', '#/resources/r2', Result.FAILED,
                          'resources:\n'
                          '  r2:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      r2_l1: { method: GET, path: "/dsaf" }\n'
                          '      r2_l2: { method: GET, path: "/dsaf" }\n'
                          '      r2_l3: { method: GET, path: "/dsaf" }\n'
                          '      self: { path: "/r2" }\n')

        self.check_result('C0303', '#/resources/r3', Result.PASSED,
                          'resources:\n'
                          '  r3:\n'
                          '    type: object\n'
                          '    links:\n'
                          '      self: { path: "/r3" }\n'
                          '      r3_l1: { method: GET, path: "/dsaf" }\n'
                          '      r3_l2: { method: GET, path: "/dsaf" }\n'
                          '      r3_l3: { method: GET, path: "/dsaf" }\n')

    def test_rule_C0200(self):
        ''' A type must have a valid description field '''

        self.check_result('C0200', '#/types/a_type', Result.PASSED,
                          'types:\n'
                          '  a_type:\n'
                          '    type: object\n'
                          '    description: A good description\n')

        self.check_result('C0200', '#/types/a_type', Result.FAILED,
                          'types:\n'
                          '  a_type:\n'
                          '    type: object\n'
                          '    description: a bad description\n')

    def test_rule_C0300(self):
        ''' A resource must have a valid description field '''

        self.check_result('C0300', '#/resources/res', Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    description: A good description\n')

        self.check_result('C0300', '#/resources/res', Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n'
                          '    description: a bad description\n')

    def test_rule_C0301(self):
        '''  A resource should be an object '''

        self.check_result('C0301', '#/resources/res', Result.PASSED,
                          'resources:\n'
                          '  res:\n'
                          '    type: object\n')

        self.check_result('C0301', '#/resources/res', Result.FAILED,
                          'resources:\n'
                          '  res:\n'
                          '    type: string\n')

    def test_rule_W0006(self):
        """ A required property should not have a default value """
        self.check_result('W0006', '#/resources/foo', Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '    required: [id]')
        self.check_result('W0006', '#/resources/foo', Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number, default: 0 }\n'
                          '    links:\n'
                          '      self: { path: /foo }\n'
                          '    required: [id]')
        # check the recursive
        self.check_result('W0006', '#/resources/foo/properties/outer',
                          Result.PASSED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '      outer:\n'
                          '        type: object\n'
                          '        properties:\n'
                          '          inner: { type: number }\n'
                          '        required: [inner]\n'
                          '    links:\n'
                          '      self: { path: /foo }\n')

        self.check_result('W0006', '#/resources/foo/properties/outer',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '      outer:\n'
                          '        type: object\n'
                          '        properties:\n'
                          '          inner: { type: number, default: 0 }\n'
                          '        required: [inner]\n'
                          '    links:\n'
                          '      self: { path: /foo }\n')

        self.check_result('W0006', '#/resources/foo/properties/outer/items',
                          Result.FAILED,
                          'resources:\n'
                          '  foo:\n'
                          '    type: object\n'
                          '    properties:\n'
                          '      id: { type: number }\n'
                          '      outer:\n'
                          '        type: array\n'
                          '        items:\n'
                          '           type: object\n'
                          '           properties:\n'
                          '              inner: {type: number,default: 0}\n'
                          '           required: [inner]\n'
                          '    links:\n'
                          '      self: { path: /foo }\n')


if __name__ == '__main__':
    logging.basicConfig(filename='test.log', level=logging.DEBUG)
    unittest.main()



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from collections import OrderedDict

from yaml.error import Mark


def obj_key_node(obj, prop):
    idx = obj.keys().index(prop)
    return obj.keys()[idx]


def add_marks_to_node(obj, start_mark=None, end_mark=None):
    if start_mark is not None:
        obj.start_mark = start_mark
    else:
        obj.start_mark = Mark(None, None, 0, 0, None, None)
    if end_mark is not None:
        obj.end_mark = end_mark
    else:
        obj.end_mark = obj.start_mark


def create_node_class(cls):
    class node_class(cls):
        def __init__(self, x, start_mark=None, end_mark=None):
            cls.__init__(self, x)
            add_marks_to_node(self, start_mark, end_mark)

        def __new__(self, x=None, start_mark=None, end_mark=None):
            return cls.__new__(self, x)
    node_class.__name__ = '%s_node' % cls.__name__
    return node_class

dict_node = create_node_class(dict)
ordered_dict_node = create_node_class(OrderedDict)
list_node = create_node_class(list)
unicode_node = create_node_class(str)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

"""
This module implements relative JSON pointers that take the form '<num>/<jsonpointer>'.
A relative JSON pointer is resolved against a data object and a base pointer.
"""

import re
import jsonpointer
from jsonpointer import JsonPointer, JsonPointerException


num_re = re.compile('^[0-9]+$')
num_hash_re = re.compile('^[0-9]+#$')
num_rel_re = re.compile('^[0-9]+/')

class RelJsonPointer(JsonPointer):
    def __init__(self, basepointer, relpointer):
        if basepointer is None:
            basepointer = ''
        super(RelJsonPointer, self).__init__(basepointer)

        self.isHash = False

        if num_re.match(relpointer):
            uplevels = int(relpointer)
            relparts = []

        elif num_hash_re.match(relpointer):
            uplevels = int(relpointer[:-1])
            relparts = []
            self.isHash = True

        elif num_rel_re.match(relpointer):
            (uplevels, relpath) = relpointer.split('/', 1)
            uplevels = int(uplevels)

            relparts = JsonPointer('/' + relpath).parts

        else:
            raise JsonPointerException(
                "Invalid relative JSON pointer '%s', " % relpointer)

        if uplevels > 0:
            if uplevels > len(self.parts):
                raise JsonPointerException(
                    "Base pointer '%s' is not deep enough for "
                    "relative pointer '%s' levels" % (basepointer, relpointer))
            self.parts = self.parts[0:-uplevels]

        if self.isHash and len(self.parts) == 0:
            raise JsonPointerException(
                "Cannot use '#' at root of relative JSON pointer '%s', "
                % relpointer)

        self.parts.extend(relparts)

    def resolve(self, doc, default=jsonpointer._nothing):
        if self.isHash:
            if len(self.parts) == 1:
                refdata = doc
            else:
                p = JsonPointer('/' + '/'.join(self.parts[:-1]))
                refdata = p.resolve(doc)
            if isinstance(refdata, list):
                return int(self.parts[-1])
            else:
                return self.parts[-1]
        else:
            return super(RelJsonPointer, self).resolve(doc, default)

def resolve_rel_pointer(doc, pointer, relpointer, default=jsonpointer._nothing):
    """
    Resolves a relative pointer against doc and pointer

    >>> obj = {'foo': {'anArray': [ {'prop': 44}], 'another prop': {'baz': 'A string' }}}

    >>> resolve_rel_pointer(obj, '/foo', '1') == obj
    True

    >>> resolve_rel_pointer(obj, '/foo/anArray', '1') == obj['foo']
    True

    >>> resolve_rel_pointer(obj, '/foo/anArray', '1/another%20prop') == obj['foo']['another prop']
    True

    >>> resolve_rel_pointer(obj, '/foo/anArray', '1/another%20prop/baz') == obj['foo']['another prop']['baz']
    True

    >>> resolve_pointer(obj, '/foo/anArray/0') == obj['foo']['anArray'][0]
    True

    """

    op = RelJsonPointer(pointer, relpointer)
    return op.resolve(doc, default)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

"""
This module defines classes for handling <json-schema> style declarations.
The `Schema` class is the base for all type-specific class such as String,
Number, Object, etc.

The normal use case is to call Schema.parse() to generate a Schema object
from Python data structure.

Consider the following:

    >>> bookdef = {
        'type': 'object',
        'description': 'Book object',
        'properties' : {
           'id' : { 'type' : 'number' },
           'title' : { 'type' : 'string' },
           'author_ids' : {
              'type' : 'array',
              'items' : { 'type' : 'number' }
           }
        }

    >>> bookschema = Schema.parse(bookdef, name='book')

    >>> print bookschema.str_simple()
    book                           object               Book object
    book.author_ids                array
    book.author_ids.[items]        number
    book.id                        number
    book.title                     string

Note that 'Array' and 'Object' schemas link to children schemas
representing the array items and object properties respectively:

    >>> type(bookschema)
    reschema.jsonschema.Object

    >>> type(bookschema['id'])
    reschema.jsonschema.Number

    >>> type(bookschema['author_ids'])
    reschema.jsonschema.Array

    >>> type(bookschema['author_ids'][0])
    reschema.jsonschema.Number

The `validate` method will ensure that input data conforms to
a schema.  This method may be called on any Schema variant, and
it recursively validates the entire input data object as needed:

    >>> bookinput = {'id': 1, 'title': 'A good book!'}
    >>> bookschema.validate(bookinput)

    >>> bookinput = {'id': 1, 'title': 'A good book!',
                     'author_ids': ['bad author']}
    >>> bookschema.validate(bookinput)
    ValidationError: 'bad author' expected to be a number for
    book.author_ids.[items]

See each individual Schema type for the complete list of validation rules.
"""

import re
import copy
import logging
import xml.etree.ElementTree as ET
from collections import OrderedDict

import uritemplate
from jsonpointer import resolve_pointer, JsonPointer

from reschema.jsonmergepatch import json_merge_patch
from reschema.parser import Parser
from reschema.util import check_type, uritemplate_required_variables, \
    uritemplate_add_query_params
from reschema.reljsonpointer import resolve_rel_pointer, JsonPointerException
from reschema.exceptions import \
    ValidationError, MissingParameter, ParseError, InvalidReference

__all__ = ['Schema']

logger = logging.getLogger(__name__)

DEFAULT_REQ_RESP = {'type': 'null'}

# Map of 'json-schema' type to class that handles it
type_map = {}


def _register_type(cls):
    type_map[cls._type] = cls


class Entity(object):
    """Base for various classes using ids and names

    :param id: URI of this entity relative to its servicedef URI

    :param name: a label for this entity, typically resembling a
        the name of a schema with its section (types, resources,
        params, etc.) with dotted or brackted qualifiers for sub-schemas
        in the style of python member accesses or array indices or
        dictionary keys.  It does not need to be unique within
        the service definition.

    :param parent: allows nesting of entities, may be None

    :param servicedef: the servicedef instance this entity is
        assocaited with. If None, inherit from parent.

    :param intermediary: a word to insert between the parent and
        instance name in the dotted fullname notation.  Typically
        used by non-Schema entities.

    :param input: parsing data relevant for error reporting, if any.
    """
    def __init__(self, id, name=None, parent=None, servicedef=None,
                 intermediary=None, input=None):
        self.id = id
        self.name = name
        self.parent = parent
        self.servicedef = servicedef

        if servicedef is None:
            if parent is None:
                # This is needed for repr(), which can end up being
                # called while printing information about the
                # ValidationError's stack trace.
                self._absolute_id = '<unspecified service uri>%s' % self.id
                self._fullname = name if name else '<unnamed>'
                raise ValidationError(
                    "Must specify 'servicedef' if parent is None", input)
            self.servicedef = parent.servicedef

        self._absolute_id = '%s%s' % (self.servicedef.id, self.id)

        if self.parent:
            if self.name is None:
                self._fullname = self.parent.fullname()
            elif intermediary is not None:
                self._fullname = '.'.join((self.parent.fullname(),
                                           intermediary, self.name))
            elif isinstance(self.parent, Array):
                self._fullname = '%s[%s]' % (self.parent.fullname(), self.name)
            elif self.parent.fullname() is None:
                self._fullname = self.name
            else:
                self._fullname = '.'.join((self.parent.fullname(), self.name))

        elif self.parent is None and self.name is None:
            self._fullname = None
        else:
            self._fullname = self.name

    def __repr__(self):
        return "<servicedef.%s '%s'>" % (self.__class__.__name__,
                                         self.fullid())

    def fullname(self):
        """Return the full printable name using dotted notation."""
        # Note: Could be property, but is a method for historical reasons.
        return self._fullname

    def fullid(self, relative=False):
        """Return the full id (canonical URI) using path notation.

        :param relative: set to True to return an id relative to this
            servicedef
        """
        # Note: Could be property, but is a method for historical reasons.
        return self.id if relative else self._absolute_id


class Schema(Entity):
    """Base class for all JSON schema types.

    Construction happens through parsing of input.  The general flow is:

        * Receive a Parser instance that is already attached to the input.
        * Parse all remaining keys at this level.
        * For keys that have a schema value, invoke Schema.parse()
        * For keys that have other values, hand the value off to the
          appropriate class without passing in the Parser.  Those classes
          will call Schema.parse() as needed on sub-fields in order
          to continue the traversal.

    Note that individual Parser instances never work with more than one
    level of input keys at a time.  Traversal happens via passing
    input sub-dictionaries back to `Schema.parse()`, which creates
    the next level of Parser down.

    :param typestr: the <json-schema> type

    :param parser: the parser of input data

    :param parent: allows nesting of schemas, may be None

    :param name: a label for this schema, typically resembling
        the name of a schema with its section (types, resources,
        params, etc.) with dotted or bracketed qualifiers for sub-schemas
        in the style of python member accesses or array indices or
        dictionary keys.  It does not need to be unique within
        the service definition.

    :param servicedef: the servicedef instance this instance is
        assocaited with. If None, inherit from parent.

    :param id: URI of this schema relative to its servicedef URI

    :raises ValidationError: if neither `servicedef` nor `parent`
        is specified.

    :raises ParseError: if unexpected data or formats are
        encountered while parsing.
    """

    # Counter used for assigning names/ids for anonymous types
    count = 1

    # Map of all known schemas by id
    schemas = {}

    def __init__(self, typestr, parser, name=None,
                 parent=None, servicedef=None, id=None):

        super(Schema, self).__init__(id=id, name=name, input=parser.input,
                                     servicedef=servicedef, parent=parent)

        self._typestr = typestr
        self.children = []

        # Save the original input object that was parsed, other
        # references may want this later.
        #
        # If used elsewhere, it *must* be considered read only.
        # Make a copy if the input needs to be changed
        self.input = parser.input

        # Give the Parser context now that this object is created
        # for setting attributes and logging messages
        parser.set_context(self.fullname(), self)

        if not self.is_ref():
            parser.parse('label', name, types=[str, unicode])
            parser.parse('description', '', types=[str, unicode])
            parser.parse('notes', '', types=[str, unicode])
            parser.parse('example')
            readOnlyDef = (parent.readOnly
                           if (parent and isinstance(parent, Schema))
                           else False)
            parser.parse('readOnly', readOnlyDef, types=bool)
            parser.parse('tags', {}, types=dict)
            parser.parse('xmlTag')
            parser.parse('xmlSchema')
            parser.parse('xmlExample')
            parser.parse('xmlKeyName')

            self.relations = OrderedDict()
            for key, value in parser.parse('relations', {},
                                           types=dict, save=False).iteritems():
                check_type(key, value, dict)
                self.relations[key] = Relation(value, key, self,
                                               id=('%s/relations/%s' %
                                                   (self.id, key)))

            self.links = OrderedDict()
            links = parser.parse('links', {}, types=dict, save=False)
            links_keys = Link.order_link_keys(links.keys())

            for key in links_keys:
                value = links[key]
                check_type(key, value, dict)
                self.links[key] = Link(value, key, self,
                                       id='%s/links/%s' % (self.id, key))

            self.anyof = []
            for i, subinput in enumerate(parser.parse('anyOf', [],
                                                      types=list, save=False)):
                s = Schema.parse(subinput, parent=self, name='anyOf[%d]' % i,
                                 id='%s/anyOf/%d' % (self.id, i))
                self.anyof.append(s)
                self.children.append(s)

            self.allof = []
            for i, subinput in enumerate(parser.parse('allOf', [],
                                                      types=list, save=False)):
                s = Schema.parse(subinput, parent=self, name='allOf[%d]' % i,
                                 id='%s/allOf/%d' % (self.id, i))
                self.allof.append(s)
                self.children.append(s)

            self.oneof = []
            for i, subinput in enumerate(parser.parse('oneOf', [],
                                                      types=list, save=False)):
                s = Schema.parse(subinput, parent=self, name='oneOf[%d]' % i,
                                 id='%s/oneOf/%d' % (self.id, i))
                self.oneof.append(s)
                self.children.append(s)

            n = parser.parse('not', save=False)
            if n is not None:
                self.not_ = Schema.parse(n, parent=self, name='not',
                                         id='%s/not' % self.id)
                self.children.append(self.not_)
            else:
                self.not_ = None

        self.schemas[self.fullid()] = self

    @classmethod
    def parse(cls, input, name=None, parent=None, servicedef=None,
              id=None):
        """Parse a <json-schema> definition for an object.

        The general flow of Schema parsing is as follows:

            * Create a Parser for the current input keys.
            * Parse enough fields to determine the Schema subclass.
            * Hand the parser and remaining input off to the constructor.

        See the constructor docstring for further parsing flow documentation.

        :param dict input: the definition to parse

        :param parent: allows nesting of schemas, may be None

        :param name: a label for this schema, typically resembling
            the name of a schema with its section (types, resources,
            params, etc.) with dotted or bracketed qualifiers for sub-schemas
            in the style of python member accesses or array indices or
            dictionary keys.  It does not need to be unique within
            the service definition.

        :param id: URI of this schema relative to its servicedef URI

        :param servicedef: the servicedef instance this instance is
            assocaited with. If None, inherit from parent.

        :raises ParseError: if unexpected data or formats are
            encountered while parsing.
        """

        if not isinstance(input, dict):
            raise ParseError(
                "Schema definition must be an object: %s%s" %
                ((parent.fullname() + '.') if parent else '', name),
                input)

        with Parser(input, name) as parser:
            if name is None:
                name = parser.parse('label', types=[str, unicode], save=False)
                if name is None:
                    name = 'element%d' % cls.count
                    cls.count = cls.count + 1

            if '$ref' in input:
                typestr = '$ref'
            elif '$merge' in input:
                typestr = '$merge'
            else:
                typestr = parser.parse('type', 'multi',
                                       types=[str, unicode], save=False)

            try:
                cls = type_map[typestr]
            except KeyError:
                msg = ('Unknown type: %s while parsing %s%s' %
                       (typestr, (parent.fullname() + '.') if parent else '',
                        name))
                raise ParseError(msg, typestr)

            return cls(parser, name, parent, servicedef=servicedef, id=id)

    @classmethod
    def find_by_id(cls, id):
        """Find a schema by fullid."""

        if id in cls.schemas:
            return cls.schemas[id]

        return None

    @property
    def typestr(self):
        """The <json-schema> type for this schema."""
        return self._typestr

    def is_simple(self):
        """Returns True if this object is a simple data type.

        Simple data types have no linked children schema.  For example, Array
        and Object return False whereas String and Number return True.

        """
        return True

    def is_ref(self):
        """Return True if this schema is a reference."""
        return False

    def is_multi(self):
        """Return True if this schema is a multi instance.

        Multi data types use the anyOf, oneOf or allOf properties
        to combine multiple schema definitions."""
        return False

    def matches(self, other):
        """ Return True if other refers to the same schema based on 'self'. """
        return (('self' in self.links) and
                ('self' in other.links) and
                (self.links['self'].path.template ==
                 other.links['self'].path.template))

    def str_simple(self):
        """Return a string representation of this element as a basic table."""
        s = '%-30s %-20s %s\n' % (self.fullname(),
                                  self.typestr,
                                  self.description.split('\n')[0])
        for child in self.children:
            s += child.str_simple()
        for relation, value in self.relations.iteritems():
            s += value.str_simple()
        return s

    def str_detailed(self, additional_details=''):
        """Return a detailed string representation of this element."""
        s = self.fullname() + ':' + self.typestr + '\n'
        if self.description:
            s += 'Description: ' + self.description + '\n'

        if additional_details:
            s += additional_details

        s += '\n'

        for child in self.children:
            s += child.str_detailed()

        return s

    def validate(self, input):
        # Must validate every schema in the allOf array
        for s in self.allof:
            s.validate(input)

        # Must validate only one schema in the oneOf array
        if len(self.oneof) > 0:
            found = 0
            for s in self.oneof:
                try:
                    s.validate(input)
                    found = found + 1
                except ValidationError:
                    continue

            if found == 0:
                raise ValidationError(
                    "%s: input does not match any 'oneOf' schema" %
                    self.fullname(), self)
            elif found > 1:
                raise ValidationError(
                    "%s: input matches more than one 'oneOf' schemas",
                    self.fullname(), self)

        # Must validate at least one schema in the anyOf array
        if len(self.anyof) > 0:
            for s in self.anyof:
                try:
                    s.validate(input)
                except ValidationError:
                    continue
                return

            raise ValidationError(
                "%s: input does not match any 'anyOf' schema" %
                self.fullname(), self)

        # Must *not* validate the not schema
        if self.not_ is not None:
            try:
                self.not_.validate(input)
                valid = True
            except ValidationError:
                valid = False

            if valid:
                raise ValidationError(
                    "%s: input should not match 'not' schema" %
                    self.fullname(), self)

    def _pointer_part_to_index(self, part):
        # Subclasses can overried to type convert if needed.
        return part

    def by_pointer(self, pointer):
        """Index into a schema by breaking a data-based jsonpointer into parts.

        Appling this method to data that does not validate against this
        schema produces undefined results.

        :param pointer: The JSON pointer.  May be either absolute or relative.
        """
        if pointer in ('/', '0'):
            # Special case root jsonpointer or relative pointer to self
            return self

        if pointer[0] == '/':
            # Absolute but non-root jsonpointer
            p = JsonPointer(pointer)
            index = self._pointer_part_to_index(p.parts[0])
            return self[index].by_pointer('/%s' % '/'.join(p.parts[1:]))

        m = re.match('^([0-9]+)(/.*)$', pointer)
        if m:
            # Looks like a relative jsonpointer
            uplevels = int(m.group(1))
            base_pointer = m.group(2)
            p = JsonPointer(base_pointer)
            o = self
            for i in range(uplevels):
                if o.parent is None:
                    raise KeyError(
                        ("%s cannot resolve '%s' as a relative JSON pointer, "
                         "not enough uplevels") % (self.fullname(), pointer))
                o = o.parent

            if len(p.parts) == 1 and p.parts[0] == '':
                return o
            return o.by_pointer(base_pointer)

        # TODO: Does this still make sense?
        #       Or should it be ValueError for json-pointer syntax?
        raise KeyError(pointer)

    def toxml(self, input, parent=None):
        """Generate an XML Element structure representing this element."""
        if parent is not None:
            parent.set(self.name, str(input))
            return parent
        else:
            elem = ET.Element(self.name)
            elem.text = str(input)
            return elem


class Multi(Schema):
    _type = 'multi'

    def __init__(self, parser, name, parent, **kwargs):
        super(Multi, self).__init__(Multi._type, parser, name, parent,
                                    **kwargs)

    def __getitem__(self, name):
        # TODO: The previous implementation, in addition to ignoring the
        #       allOf and oneOf lists, was highly order-dependent.
        #       Factoring out by_pointer() somehow exposed this sufficiently
        #       to fail the unit tests (it is not clear how they were passing
        #       before), so for now call this not implemented and xfail
        #       the tests- consistent failure is better than
        #       non-deterministic failure.
        raise NotImplementedError

    def is_multi(self):
        return True

    @property
    def typestr(self):
        if len(self.children) == 0:
            return "any"
        return "multiple"

_register_type(Multi)


class DynamicSchema(Schema):

    def __init__(self, type, parser, name, parent, **kwargs):
        super(DynamicSchema, self).__init__(type, parser, name, parent,
                                            **kwargs)

    @property
    def typestr(self):
        return self.refschema.name

    def is_simple(self):
        return False

    def is_ref(self):
        """Return True if this schema is a reference."""
        return True

    def validate(self, input):
        self.refschema.validate(input)

    def toxml(self, input, parent=None):
        return self.refschema.toxml(input, parent)

    def __getitem__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        return self.refschema.__getitem__(name)

    def __getattr__(self, name):
        # Guard against infinite recursion - need to check both
        # 'reschema' and '_reschema' because the latter is used
        # by the property 'reschema'
        if name == 'refschema' or name == '_refschema':
            raise AttributeError()
        return getattr(self.refschema, name)

    def by_pointer(self, pointer):
        """
        Index into a schema by breaking a data-based jsonpointer into parts.
        """
        return self.refschema.by_pointer(pointer)


class Ref(DynamicSchema):
    _type = '$ref'

    def __init__(self, parser, name, parent, **kwargs):
        super(Ref, self).__init__(Ref._type, parser, name, parent, **kwargs)

        # Lazy resolution because references may be used before they
        # are defined
        self._refschema = None
        parser.parse('$ref', required=True, save_as='_refschema_id')
        if len(parser.input.keys()) != 1:
            raise ParseError("$ref object may not have any other properties",
                             parser.input)

    @property
    def refschema(self):
        if self._refschema is None:
            sch = Schema.find_by_id(self._refschema_id)
            if sch is None:
                sch = self.servicedef.find(self._refschema_id)
            if sch is None:
                raise InvalidReference(("%s $ref" % self.fullname()),
                                       self._refschema_id)

            self._refschema = sch
            for link in sch.links:
                if link not in self.links:
                    self.links[link] = sch.links[link]
            for relation in sch.relations:
                if relation not in self.relations:
                    self.relations[relation] = sch.relations[relation]

        return self._refschema

    @property
    def typestr(self):
        return self.refschema.name

    def is_simple(self):
        return False

    def is_ref(self):
        """Return True if this schema is a reference."""
        return True

    def validate(self, input):
        self.refschema.validate(input)

    def toxml(self, input, parent=None):
        return self.refschema.toxml(input, parent)

    def __getitem__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        return self.refschema.__getitem__(name)

    def __getattr__(self, name):
        # Guard against infinite recursion - need to check both
        # 'reschema' and '_reschema' because the latter is used
        # by the property 'reschema'
        if name == 'refschema' or name == '_refschema':
            raise AttributeError()
        return getattr(self.refschema, name)

    def by_pointer(self, pointer):
        """
        Index into a schema by breaking a data-based jsonpointer into parts.
        """
        return self.refschema.by_pointer(pointer)

_register_type(Ref)


class Merge(DynamicSchema):
    _type = '$merge'

    def __init__(self, parser, name, parent, **kwargs):
        super(Merge, self).__init__(Merge._type, parser, name, parent,
                                    **kwargs)

        # Lazy resolution because references may be used before they
        # are defined
        merge = parser.parse('$merge', required=True, save=False)
        with Parser(merge, name, self) as merge_parser:
            merge_parser.parse('source', save_as='_mergesource', required=True)
            merge_parser.parse('with', save_as='_mergewith', required=True)

        self._refschema = None

    @property
    def refschema(self):
        if self._refschema is None:
            merged = json_merge_patch(self.servicedef,
                                      self._mergesource, self._mergewith)

            self._refschema = Schema.parse(merged, name=self.name,
                                           servicedef=self.servicedef,
                                           id=self.id)
            # Need to assign parent *after* assigning _refschema,
            # because otherwise we hit a recursion when doing some
            # logging that would try to resolve _refschema again...
            self._refschema.parent = self

        return self._refschema

_register_type(Merge)


class Null(Schema):
    _type = 'null'

    def __init__(self, parser, name, parent, **kwargs):
        super(Null, self).__init__(Null._type, parser, name, parent, **kwargs)

    def validate(self, input):
        if (input is not None):
            raise ValidationError("%s should be None, got '%s'" %
                                  (self.fullname(), type(input)), self)
        super(Null, self).validate(input)

_register_type(Null)


class Boolean(Schema):
    _type = 'boolean'

    def __init__(self, parser, name, parent, **kwargs):
        super(Boolean, self).__init__(Boolean._type, parser, name, parent,
                                      **kwargs)
        parser.parse('default')
        parser.parse('enum')

    def validate(self, input):
        if (type(input) is not bool):
            raise ValidationError("%s should be a boolean, got '%s'" %
                                  (self.fullname(), type(input)), self)
        if (self.enum is not None) and (input not in self.enum):
            raise ValidationError(
                "%s: input not a valid enumeration value: %s" %
                (self.fullname(), input), self)

        super(Boolean, self).validate(input)

_register_type(Boolean)


class String(Schema):
    _type = 'string'

    def __init__(self, parser, name, parent, **kwargs):
        super(String, self).__init__(String._type, parser, name, parent,
                                     **kwargs)
        parser.parse('minLength', types=int)
        parser.parse('maxLength', types=int)
        parser.parse('pattern')
        parser.parse('enum')
        parser.parse('default')

    def validate(self, input):
        def trunc():
            if len(unicode(input)) > 40:
                return unicode(input)[:40] + "..."
            else:
                return unicode(input)

        if not isinstance(input, (str, unicode)):
            raise ValidationError("%s: input must be a string, got %s: %s" %
                          (self.fullname(), type(input), trunc()), self)

        if (self.minLength is not None) and len(input) < self.minLength:
            raise ValidationError(
                "%s: input must be at least %d chars, got %d: %s" %
                (self.fullname(), self.minLength, len(input), trunc), self)

        if (self.maxLength is not None) and len(input) > self.maxLength:
            raise ValidationError(
                "%s: input must be no more than %d chars, got %d: %s" %
                (self.fullname(), self.maxLength, len(input), trunc), self)

        if (self.pattern is not None) and (not re.match(self.pattern, input)):
            raise ValidationError(
                "%s: input failed pattern match %s: %s" %
                (self.fullname(), self.pattern, trunc), self)

        if (self.enum is not None) and (input not in self.enum):
            raise ValidationError(
                "%s: input not a valid enumeration value: %s" %
                (self.fullname(), trunc), self)
        super(String, self).validate(input)

    def str_detailed(self):
        s = ''
        if self.minLength or self.maxLength:
            s += 'Length:'
            if self.minLength:
                s += ' min %d' % self.minLength
            if self.maxLength:
                s += ' max %d' % self.maxLength
            s += '\n'
        if self.pattern:
            s += 'Pattern: %s\n' % self.pattern
        if self.enum:
            s += 'Valid values: %s\n' % self.enum
        if self.default:
            s += 'Default value: %s\n' % self.default
        return super(String, self).str_detailed(additional_details=s)
_register_type(String)


class NumberOrInteger(Schema):
    _type = 'number'

    def __init__(self, type, allowed_types, parser, name, parent, **kwargs):
        super(NumberOrInteger, self).__init__(type, parser,
                                              name, parent, **kwargs)
        self.allowed_types = allowed_types
        parser.parse('minimum', types=allowed_types)
        parser.parse('maximum', types=allowed_types)
        parser.parse('exclusiveMinimum', False, types=bool)
        parser.parse('exclusiveMaximum', False, types=bool)
        parser.parse('default', types=allowed_types)
        parser.parse('enum', types=list)

    def validate(self, input):
        if (not any(isinstance(input, t) for t in self.allowed_types) or
                isinstance(input, bool)):
            raise ValidationError("%s should be a number, got '%s'" %
                                  (self.fullname(), type(input)), self)

        if self.minimum is not None:
            if self.exclusiveMinimum:
                if not (input > self.minimum):
                    raise ValidationError(
                        "%s: input must be > minimum %d, got %d" %
                        (self.fullname(), self.minimum, input), self)
            else:
                if not (input >= self.minimum):
                    raise ValidationError(
                        "%s: input must be >= minimum %d, got %d" %
                        (self.fullname(), self.minimum, input), self)

        if self.maximum is not None:
            if self.exclusiveMaximum:
                if not (input < self.maximum):
                    raise ValidationError(
                        "%s: input must be < maximum %d, got %d" %
                        (self.fullname(), self.maximum, input), self)
            else:
                if not (input <= self.maximum):
                    raise ValidationError(
                        "%s: input must be <= maximum %d, got %d" %
                        (self.fullname(), self.maximum, input), self)

        if (self.enum is not None) and (input not in self.enum):
            raise ValidationError(
                "%s: input not a valid enumeration value: %s" %
                (self.fullname(), input), self)
        super(NumberOrInteger, self).validate(input)

    def str_detailed(self):
        s = ''
        if self.minimum or self.maximum:
            s += 'Range: '
            if self.minimum is not None:
                s += ' min %d' % self.minimum
            if self.maximum is not None:
                s += ' max %d' % self.maximum
            s += '\n'
        if self.enum:
            s += 'Valid values: %s\n' % self.enum
        if self.default:
            s += 'Default value: %s\n' % self.default
        return super(NumberOrInteger, self).str_detailed(additional_details=s)


class Number(NumberOrInteger):
    _type = 'number'

    def __init__(self, parser, name, parent, **kwargs):
        super(Number, self).__init__(self._type, (int, float, long), parser,
                                     name, parent, **kwargs)

_register_type(Number)


class Integer(NumberOrInteger):
    _type = 'integer'

    def __init__(self, parser, name, parent, **kwargs):
        super(Integer, self).__init__(self._type, (int, long), parser,
                                      name, parent, **kwargs)

_register_type(Integer)


class Timestamp(Schema):
    _type = 'timestamp'

    def __init__(self, parser, name, parent, **kwargs):
        super(Timestamp, self).__init__(Timestamp._type, parser, name, parent,
                                        **kwargs)

    def validate(self, input):
        if (not any(isinstance(input, t) for t in (int, float, long)) or
                isinstance(input, bool)):
            raise ValidationError("'%s' expected to be a number for %s" %
                                  (input, self.fullname()), self)
        super(Timestamp, self).validate(input)

_register_type(Timestamp)


class TimestampHP(Schema):
    _type = 'timestamp-hp'

    def __init__(self, parser, name, parent, **kwargs):
        super(TimestampHP, self).__init__(TimestampHP._type, parser,
                                          name, parent, **kwargs)

    def validate(self, input):
        if (not any(isinstance(input, t) for t in (int, float, long)) or
                isinstance(input, bool)):
            raise ValidationError("'%s' expected to be a number for %s" %
                                  (input, self.fullname()), self)
        super(TimestampHP, self).validate(input)

_register_type(TimestampHP)


class Object(Schema):
    _type = 'object'

    def __init__(self, parser, name, parent, **kwargs):
        super(Object, self).__init__(Object._type, parser, name, parent,
                                     **kwargs)
        self.properties = OrderedDict()

        for prop, value in parser.parse('properties', {},
                                        types=dict, save=False).iteritems():
            c = Schema.parse(value, prop, self,
                             id='%s/properties/%s' % (self.id, prop))
            self.properties[prop] = c
            self.children.append(c)

        parser.parse('required', types=[list])

        ap = parser.parse('additionalProperties',
                          types=[dict, bool], save=False)
        if ap in (None, True):
            ap = {}
        if type(ap) is bool:
            self.additional_properties = ap
        else:
            ap_name = ap['label'] if ('label' in ap) else 'prop'
            c = Schema.parse(ap, '<%s>' % ap_name, self,
                             id='%s/additionalProperties' % (self.id))
            self.additional_properties = c
            self.children.append(c)

    def __getitem__(self, name):
        if name in self.properties:
            return self.properties[name]
        if self.additional_properties is not False:
            return self.additional_properties
        # Be lazy about generating the right kind of exception:
        self.properties[name]

    def is_simple(self):
        return False

    def validate(self, input):
        if not isinstance(input, dict):
            raise ValidationError("%s should be an object, got '%s'" %
                                  (self.fullname(), type(input)), self)

        for k in input:
            if k in self.properties:
                self.properties[k].validate(input[k])
            elif self.additional_properties is False:
                raise ValidationError("'%s' is not a valid property for %s" %
                                      (k, self.fullname()), self)
            elif isinstance(self.additional_properties, Schema):
                self.additional_properties.validate(input[k])

        if self.required is not None:
            for k in self.required:
                if k not in input:
                    raise ValidationError(
                        "Missing required property '%s' for '%s'" %
                        (k, self.fullname()), self)
        super(Object, self).validate(input)

    def toxml(self, input, parent=None):
        """Return ElementTree object with `input` data.

        Additional Properties that are not explicitly defined are not
        supported.

        """
        if parent is not None:
            elem = ET.SubElement(parent, self.name)
        else:
            elem = ET.Element(self.name)

        for k in input:
            if k not in self.properties:
                if self.additional_properties is False:
                    raise ValueError('Invalid property: %s' % k)

                subobj = copy.copy(self.additional_properties)
                keyname = subobj.xmlKeyName or 'key'

                if subobj.is_simple():
                    subelem = ET.SubElement(elem, subobj.name)
                    subelem.set(keyname, k)

                    subelem.text = str(input[k])
                else:
                    subelem = subobj.toxml(input[k])
                    subelem.set(keyname, k)
                    elem.append(subelem)

            else:
                prop = self.properties[k]
                prop.toxml(input[k], elem)

        return elem
_register_type(Object)


class Array(Schema):
    _type = 'array'

    def __init__(self, parser, name, parent, **kwargs):
        super(Array, self).__init__(Array._type, parser, name, parent,
                                    **kwargs)

        items = parser.parse('items', required=True, save=False)
        if 'label' in items:
            childname = items['label']
        else:
            childname = 'items'
        self.items = Schema.parse(items, childname, self,
                                  id='%s/items' % self.id)
        self.children.append(self.items)

        parser.parse('minItems')
        parser.parse('maxItems')

    def is_simple(self):
        return False

    @property
    def typestr(self):
        return 'array of <%s>' % self.items.typestr

    def __getitem__(self, name):
        # Use internal function for convenience even if not pointer part.
        # This is just to get the right exception on both code paths.
        self._pointer_part_to_index(name)
        return self.items

    def _pointer_part_to_index(self, part):
        try:
            index = int(part)
        except ValueError:
            raise TypeError("list indices must be integers, not %r" %
                            type(part).__name__)
        return index

    def validate(self, input):
        if not isinstance(input, list):
            raise ValidationError("%s should be an array, got '%s'" %
                                  (self.fullname(), type(input)), self)

        if (self.minItems is not None) and (len(input) < self.minItems):
            raise ValidationError(
                "%s: input must be at least %d items, got %d" %
                (self.fullname(), self.minItems, len(input)), self)

        if (self.maxItems is not None) and (len(input) > self.maxItems):
            raise ValidationError(
                "%s: input must be no more than %d items, got %d" %
                (self.fullname(), self.maxItems, len(input)), self)

        for o in input:
            self.items.validate(o)

        super(Array, self).validate(input)

    def toxml(self, input, parent=None):
        if parent is not None:
            elem = ET.SubElement(parent, '%s' % self.name)
        else:
            elem = ET.Element(self.name)

        for o in input:
            subelem = self.children[0].toxml(o)
            elem.append(subelem)
        return elem
_register_type(Array)


class Data(Schema):
    _type = 'data'

    def __init__(self, parser, name, parent, **kwargs):
        super(Data, self).__init__(Data._type, parser, name, parent, **kwargs)

        parser.parse('content_type', required=True)
        parser.parse('description')

    def validate(self, input):
        # any value will pass, regardless of content_type set
        # validation of that type of data seems beyond the scope of reschema
        pass

    def is_simple(self):
        return True
_register_type(Data)


class Relation(Entity):

    def __init__(self, input, name, schema, id):
        super(Relation, self).__init__(id=id, name=name, parent=schema,
                                       intermediary='relations')
        self.schema = schema
        self.vars = None

        with Parser(input, self.fullname(), self) as parser:
            # Lazy resolution because references may be used before they
            # are defined
            self._resource = None
            parser.parse('resource', required=True, save_as='_resource_id')
            parser.parse('vars')
            parser.parse('description', '')
            parser.parse('tags', {}, types=dict)

    def __str__(self):
        return self.name

    def is_ref(self):
        return False

    def is_multi(self):
        return False

    @property
    def resource(self):
        if self._resource is None:
            sch = Schema.find_by_id(self._resource_id)
            if sch is None:
                sch = self.servicedef.find(self._resource_id)
            if sch is None:
                raise InvalidReference(("%s resource" % self.fullname()),
                                       self._resource_id)
            self._resource = sch
        return self._resource

    def str_simple(self):
        return '%-30s %-20s\n' % (self.fullname(), '<relation>')

    def resolve(self, data=None, fragment='', kvs=None):
        """ Resolves this path against data and kvs.

        :param obj data: object to use for resolving relation vars
        :param str fragment: relative JSON pointer into indicating
            relative starting point within data to resolve vars
        :param obj kvs: override key / value pairs for the target
            resource's path vars

        :return: a tuple (uri, values), uri is the
           fully resolved path and values is the
           complete dictionary of all resolved values used
           to complete the path

        """
        target_self = self.resource.links['self']

        if kvs is None:
            kvs = {}
        else:
            # We're going to add vals to kvs, so copy to
            # make sure we don't tweak the callers dict
            kvs = copy.copy(kvs)

        if self.vars is not None:
            # If the resource defines 'vars', use data to
            # resolve them and put them in kvs
            for var, relp in self.vars.iteritems():
                if var in kvs:
                    # This var's value was provided directly
                    # by the caller in kvs, so take the caller's
                    # version
                    continue

                if data is None:
                    # If var was not in kvs and no data, raise an exception
                    raise MissingParameter(
                        "Missing value for relation '%s' var: %s" %
                        (str(self), var), self)

                try:
                    kvs[var] = resolve_rel_pointer(data, fragment, relp)
                except JsonPointerException:
                    raise MissingParameter(
                        ("Relation %s failed to assign var %s from data "
                         "using rel pointer %s") %
                        (self.fullname(), var, relp), self)

        (uri, values) = target_self.path.resolve(data=None, kvs=kvs)

        return (uri, values)


class Link(Entity):

    def __init__(self, input, name, schema, id):
        super(Link, self).__init__(id=id, name=name, parent=schema,
                                   intermediary='links')
        self.schema = schema

        with Parser(input, self.fullname(), self) as parser:
            parser.parse('description', '')
            parser.parse('notes', '')
            parser.parse('example')
            parser.parse('method')
            parser.parse('authorization')
            parser.parse('tags', {}, types=dict)

            pathdef = parser.parse('path', save=False)
            if pathdef is not None:
                if name == 'self':
                    # Backwards compatibility, treat 'params' as
                    # 'path.vars'
                    params = parser.parse('params', {}, save=False)
                else:
                    params = {}
                self.path = Path(self, pathdef, additional_vars=params)
                parser.mark_object(self.path, 'path')
            elif self.method is not None:
                if 'self' not in self.schema.links:
                    raise ParseError(("Link '%s' defined with no path and "
                                      "schema has no 'self' link") %
                                     str(self), parser.input)
                self.path = self.schema.links['self'].path
            else:
                self.path = None

            # Everything other than 'self' requires a method set
            if (name != 'self') and not self.method:
                raise ParseError("Link '%s' does not have a method set"
                                 % str(self), parser.input)

            self._request = parser.parse('request', save=False)
            if self._request is not None:
                self._request = Schema.parse(self._request,
                                             parent=self, name='request',
                                             id='%s/request' % self.id)
            elif name != 'self':
                self._request = Schema.parse(DEFAULT_REQ_RESP,
                                             parent=self, name='request',
                                             id='%s/request' % self.id)

            self._response = parser.parse('response', save=False)
            if self._response is not None:
                self._response = Schema.parse(self._response,
                                              parent=self, name='response',
                                              id='%s/response' % self.id)
            elif name != 'self':
                self._response = Schema.parse(DEFAULT_REQ_RESP,
                                              parent=self, name='response',
                                              id='%s/response' % self.id)

    @classmethod
    def order_link_keys(self, keys):
        # Make sure to process 'self' first, other links may rely
        # on 'self.path'
        new_keys = []
        for k in keys:
            if k == 'self':
                new_keys.append(k)
                keys.remove('self')
        new_keys.extend(keys)
        return new_keys

    def is_ref(self):
        return False

    def is_multi(self):
        return False

    def str_simple(self):
        return '%-30s %-20s %s\n' % (self.fullname(), '<link>',
                                     self.description)

    @property
    def request(self):
        r = self._request
        if type(r) is Ref:
            r = r.refschema
        return r

    @property
    def response(self):
        r = self._response
        if type(r) is Ref:
            r = r.refschema
        return r


class Path(Entity):
    """The `Path` class manages URI templates and resolves variables.

    The `resolve` method supports resolving variables in the template from
    a data object conforming to the linked schema.

    When creating a path, the `pathdef` may take one of two forms:

      1) pathdef = '<uri-template>'

      2) pathdef = { 'template': '<uri-template'>,
                     'vars' : {
                        '<var1>' : <rel-json-pointer>,
                        '<var2>' : <rel-json-pointer>,
                        ... } }

    In both cases, the '<uri-template>' is follows RFC6570 syntax.
    Variables in the template are resolved first looking to vars, then
    do the data object.

    Since URI templates can only specify simple variables, the second
    form is used as a level of indirection to move up or down in a
    data structure relative to the place where the link was defined.

    """

    def __init__(self, link, pathdef, additional_vars=None):
        """Create a `Path` associated with `link`."""

        super(Path, self).__init__(id='%s/path' % link.id,
                                   name='path',
                                   parent=link)
        self.link = link

        self.pathdef = pathdef
        self.vars = {}
        self.var_schemas = {}

        if isinstance(pathdef, dict):
            self.template = pathdef['template']
            self.vars = pathdef['vars']
        else:
            self.template = pathdef
            self.vars = {}

        if additional_vars:
            for k, v in additional_vars.iteritems():
                self.vars[k] = v

        # For any URI template parameters that are in the
        # template but not explicitly listed in vars, add them
        # with a relpath of '0/{param}'
        #
        # Example:
        #   path:
        #     template: '$/books/{id}'
        #
        # implies
        #   path:
        #     template: '$/books/{id}'
        #     vars:
        #       id: '0/id'
        #
        all = set(uritemplate.variables(self.template))
        for param in all:
            if param == '$':
                continue

            if param not in self.vars:
                self.vars[param] = ('0/%s' % param)

        # Any vars that have a target that is a dictionanry
        # as a schema.  This allows additional parameters
        # that cannot be validated or described by data
        #
        # Example:
        #   path:
        #     template: '$/books/{id}{?offset}'
        #     vars:
        #       offset: { type: integer }
        #
        for var, target in self.vars.iteritems():
            if isinstance(target, dict):
                self.var_schemas[var] = Schema.parse(
                    target, parent=self.link.schema, name=var,
                    id=('%s/vars/%s' % (self.link.schema.id, var)))
                self.vars[var] = None

        # Any vars that are listed but *not* in the URI template are
        # added to the template as optional query parmas
        #
        # Example:
        #   path:
        #     template: '$/books/{id}'
        #     vars:
        #       offset: { type: integer }
        #
        # template becomes: '$/books/{id}{?offset}'
        #
        missing = set(self.vars.keys()) - all
        if missing:
            self.template = uritemplate_add_query_params(
                self.template, set(self.vars.keys()) - all)

        # Special '$' var is replaced with the schema at the same
        # location
        self.vars['$'] = '0'
        self.var_schemas['$'] = self.link.schema

    def __str__(self):
        return self.template

    def resolve(self, data=None, pointer=None, kvs=None, validate=False):
        """Resolve variables in template from `data` relative to `pointer`.

        :param obj data: source data to use for path vars
        :param str pointer: relative JSON pointer to index into data
        :param obj kvs: override key / value pairs for template vars
        :param bool validate: if true, validate data and kvs according
            to the defined schemas

        :return: a tuple (uri, values), uri is the fully resolved
           path, and values is the complete dictionary of all resolved
           values used to complete the path

        A path object has a template with 0 or more variables in it
        that must be resolved.  The template varaiables are resolved
        first from `kvs` directly, then from data.  When data is used,
        if any path.vars are defined, they are used for resolution.
        Remaining template vars are resolved directly from data.

        Example with direct resolution:
           path: '$/foos/{id}

           data           kvs           uri
           {'id': 5}      None          $/foos/5
           {'id': 5}      {'id': 6}     $/foos/6
           None           {'id': 6}     $/foos/6

        Example with indirect resolution:
           path:
              template: '$/foos/{id}
              vars: { id: '0/sub/id' }

           data                    kvs           uri
           {'sub': {'id': 5}}      None          $/foos/5
           {'sub': {'id': 5}}      {'id': 6}     $/foos/6

        Example with indirect resolution and a fragment
           path:
              template: '$/foos/{id}
              vars: { id: '1/suba/id' }

           fragment='/subb'

           data                              kvs           uri
           {'suba': {'id': 5}, 'subb': 7}    None          $/foos/5
           {'suba': {'id': 5}, 'subb': 7}    {'id': 6}     $/foos/6

        """
        # Collect the set of required variables from the template.
        # Any uri template variables starting with ? or & are optional
        required = set(uritemplate_required_variables(self.template))

        if kvs is None:
            kvs = {}
        else:
            # We're going to add vals to kvs, so copy to
            # make sure we don't tweak the callers dict
            kvs = copy.copy(kvs)

        if data:
            # $ is a special value allowing the complete data
            # (relatively) to be inserted into the template via {$}
            kvs['$'] = resolve_pointer(data, pointer or '')

            # Evaluate path.vars from the data
            for var, relp in (self.vars or {}).iteritems():
                if var in kvs or relp is None:
                    # Skip if either this var's value was provided
                    # directly by the caller in kvs, or if there is no
                    # relative pointer and the data can *only* be provided
                    # by kvs
                    continue

                try:
                    kvs[var] = resolve_rel_pointer(data, pointer or '', relp)
                except JsonPointerException:
                    # Only fail for required params.  If the param is optional,
                    # the data may not be present so just leave out of the kvs
                    if var in required:
                        raise MissingParameter(
                            ("Path %s failed to assign var %s from data "
                             "using rel pointer %s") % (self, var, relp),
                            self)

        tmpl = self.template
        logger.debug("%s template: %s" % (self.link.fullname(), tmpl))
        have = set(kvs.keys())
        if not required.issubset(have):
            raise MissingParameter(
                "Missing parameters for link '%s' path template '%s': %s" %
                (self.link.fullname(), self.template,
                 [x for x in required.difference(have)]),
                self)

        if validate:
            for var in have:
                if var not in self.var_schemas:
                    relp = self.vars[var]
                    self.var_schemas[var] = self.link.schema.by_pointer(relp)

                self.var_schemas[var].validate(kvs[var])

        # Convert values to strings, as otherwise they might get "dropped"
        # If the value is 0, it will get dropped
        uri_kvs = {}
        for k,v in kvs.iteritems():
            uri_kvs[k] = str(v)

        uri = uritemplate.expand(self.template, uri_kvs)

        return (uri, kvs)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import copy

import pytest
import mock

from reschema.parser import Parser
from reschema.exceptions import ParseError


ID_HOSTNAME = 'http://support.riverbed.com'
ID_BASE = '%s/apis/testschema/1.0' % ID_HOSTNAME
ID_FRAGMENT = '#/foo/bar'
ID_ABSPATH = '/apis/otherschema/2.0'
ID_ABSPATH_FRAGMENT = '%s%s' % (ID_ABSPATH, ID_FRAGMENT)

INPUT_DICT = {
    'foo': 42,
    'bar': {'$ref': ID_FRAGMENT},
}

INPUT_LIST = ['foo', {'$ref': ID_FRAGMENT}]
INPUT_LIST_EXPANDED = ['foo', {'$ref': '%s%s' % (ID_BASE, ID_FRAGMENT)}]

ANY_NAME = 'whatever'
ANY_OBJ = {'stuff', 'nonsense'}


@pytest.fixture
def parser():
    p = Parser(INPUT_DICT, ANY_NAME)
    return p


def test_parser_dict_only():
    with pytest.raises(ParseError):
        Parser(INPUT_LIST, ANY_NAME)

    with pytest.raises(ParseError):
        Parser(INPUT_LIST, '{"foo": "bar"}')


def test_parser_context():
    with mock.patch('reschema.parser.Parser.set_context'):
        p = Parser(INPUT_DICT, ANY_NAME, ANY_OBJ)
        assert p.set_context.called_once_with(ANY_NAME, ANY_OBJ)


def test_set_context(parser):
    other_name = 'different_from_any_name'
    p = Parser(INPUT_DICT, ANY_NAME)
    p.mark_object = mock.Mock()
    p.set_context(other_name, ANY_OBJ)

    assert p.name == other_name
    assert p.obj is ANY_OBJ
    assert p.mark_object.called_once_with(ANY_OBJ)


def test_local_ref():
    assert Parser.expand_ref(
        ID_BASE, ID_FRAGMENT) == '%s%s' % (ID_BASE, ID_FRAGMENT)


def test_provider_ref():
    expanded = Parser.expand_ref(ID_BASE, ID_ABSPATH_FRAGMENT)
    assert expanded == '%s%s' % (ID_HOSTNAME, ID_ABSPATH_FRAGMENT)


def test_full_ref():
    id_ = '%s%s' % (ID_HOSTNAME, ID_ABSPATH_FRAGMENT)
    assert Parser.expand_ref(ID_BASE, id_) == id_


def test_expand_refs_empty_input():
    # Shouldn't raise an exception, nothing else to check.
    Parser.expand_refs(ID_BASE, None)

    input_ = {}
    Parser.expand_refs(ID_BASE, input_)
    assert input_ == {}


def test_expanded_refs_array():
    # This ends up covering the dict case as well, since the $ref must
    # be a dict itself.
    input_ = copy.deepcopy(INPUT_LIST)
    Parser.expand_refs(ID_BASE, input_)
    assert input_ == INPUT_LIST_EXPANDED


def test_preprocess():
    with mock.patch('reschema.parser.Parser.expand_refs'):
        Parser.preprocess(ID_BASE, copy.deepcopy(INPUT_LIST))
        Parser.expand_refs.assert_called_once_with(ID_BASE, INPUT_LIST)


def test_preprocess_input(parser):
    with mock.patch('reschema.parser.Parser.preprocess'):
        parser.preprocess_input(ID_BASE)
        parser.preprocess.assert_called_once_with(ID_BASE, parser.input)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import re
from reschema.exceptions import ParseError

# copy params from previous uritemplate version
RESERVED = ":/?#[]@!$&'()*+,;="
OPERATOR = "+#./;?&|!@"
MODIFIER = ":^"
TEMPLATE = re.compile("{([^\}]+)}")


def check_type(prop, val, valid_type, obj=None):
    if type(valid_type) is not list:
        valid_type = [valid_type]

    valid = False
    for t in valid_type:
        if isinstance(val, t):
            valid = True
            break

    if not valid:
        msg = ("Value provided for '%s' must be %s, got %s" %
               (prop, str(valid_type), type(val)))
        raise ParseError(msg, prop, obj)


def a_or_an(s):
    if s[0] in ('a', 'e', 'i', 'o', 'u'):
        return "an"
    else:
        return "a"


def str_to_id(s, c='_'):
    """Convert the input string to a valid HTML id by replacing all invalid
    characters with the character C."""
    return (re.sub('[^a-zA-Z0-9-:.]', c, s)).strip('_')


def uritemplate_required_variables(template):
    """Returns the set of vars in a uri template that are required.

    Query related variables are prefixed by ? or &:

       template = 'http://foo.bar/items/{id}?x={x}{&y,z}{#frag}'

    x, y, and z are considered query variables and will be skipped.

    """
    vars = set()
    for varlist in TEMPLATE.findall(template):
        if varlist[0] in "?&":
            continue
        if varlist[0] in OPERATOR:
            varlist = varlist[1:]
        varspecs = varlist.split(',')
        for var in varspecs:
            # handle prefix values
            var = var.split(':')[0]
            # handle composite values
            if var.endswith('*'):
                var = var[:-1]
            vars.add(var)
    return vars


def uritemplate_add_query_params(template, params):
    if not params:
        return template

    for varlist in TEMPLATE.findall(template):
        if varlist[0] in "?&":
            orig = varlist
            updated = orig + ',' + ','.join(params)
            return template.replace(orig, updated)

    return "%s{?%s}" % (template, ','.join(params))



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from reschema.loader_nodes import obj_key_node


#
# General exception and base class
#
class ReschemaException(Exception):
    """ Base exception class for reschema errors. """


class UnsupportedSchema(ReschemaException):
    """ Schema uses an unsupported schema format. """
    pass


class InvalidServiceId(ReschemaException):
    """ No service defintion loaded for the given id. """
    pass


class InvalidServiceName(ReschemaException):
    """ No service defintion loaded for the given name. """
    pass


class DuplicateServiceId(ReschemaException):
    """ Service defintion already registered for the requested id. """

    def __init__(self, id_):
        self.id_ = id_

    def __str__(self):
        return "Service definition already registered by id: %s" % self.id_


class NoContext(ReschemaException):
    """ A relative reference was provided with no supporting context. """

    def __init__(self, reference):
        self.reference = reference

    def __str__(self):
        return ("Relative reference cannot be resolved without context: %s"
                % self.reference)


class NoManager(ReschemaException):
    """ A partial/full reference was provided but no ServiceDefManager available. """

    def __init__(self, reference):
        self.reference = reference

    def __str__(self):
        return ("Reference cannot be resolved without a manager: %s"
                % self.reference)


class InvalidReference(ReschemaException):
    """ Invalid reference. """

    def __init__(self, msg, reference):
        self.msg = msg
        self.reference = reference

    def __str__(self):
        return ("Invalid reference '%s': %s" % (self.reference, self.msg))


#
# jsonschema parsing errors
#
class MarkedError(ReschemaException):
    """ Base exception class for marked schema objects. """
    def __init__(self, message, obj=None, parent_obj=None):
        super(MarkedError, self).__init__(message)

        if hasattr(obj, 'start_mark'):
            self.obj = obj
            self.start_mark = obj.start_mark
            self.end_mark = obj.end_mark
        elif (  parent_obj and (obj in parent_obj) and
                hasattr(parent_obj, 'start_mark')):
            self.obj = obj_key_node(parent_obj, obj)
            self.start_mark = self.obj.start_mark
            self.end_mark = self.obj.end_mark
        else:
            self.start_mark = None
            self.end_mark = None

    def __str__(self):
        msg = []
        msg.append(super(MarkedError, self).__str__())
        if self.start_mark:
            msg.append('Source data locations:')
            msg.append('Start mark: ' + str(self.start_mark))
        if self.end_mark:
            msg.append('End mark:   ' + str(self.end_mark))
        return '\n'.join(msg)


class MissingParameter(MarkedError):
    """ Missing link parameters. """


class ParseError(MarkedError):
    """ Schema parsing error. """


#
# jsonschema validation errors
#
class ValidationError(ReschemaException):
    """ Schema validation error. """


class ReschemaLoadHookException(Exception):
    """ Exceptions as a result of attempting loading via hooks. """



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

"""
This module implements JSON merge-patch according to
http://tools.ietf.org/html/draft-ietf-appsawg-json-merge-patch-02

This implementation supports resolution of $ref/$merge
only when needed.

"""

import copy
import logging
import json

from reschema.parser import Parser
from reschema.exceptions import ParseError
import reschema.settings


logger = logging.getLogger(__name__)


def isdebug():
    return (reschema.settings.VERBOSE_DEBUG and
            logger.isEnabledFor(logging.DEBUG))


def _eval_shallow(servicedef, obj, need_copy=False):
    """Resolve $ref and $merge in obj, using servicedef for remote references.

    This function takes an input object which is a dictionary, and evaluates
    $merge / $ref only if they are present at the root of the object.

    If need_copy is True, a modifyable shallow copy is returned.

    """

    # _eval_shallow() resolves $ref and $merge to their values in
    # source and with_.  This is a *shallow* evaluation in that embedded
    # $ref or $merge at deeper levels are *not* resolved.
    #
    # For example, the following will be resolved:
    #    { $ref: ... }
    #    { $merge: ... }
    #
    # But the following will *not* be resolved
    #    { type: object,
    #      properties: { x: { $ref: ... } } }
    #
    # Need to loop in the event that a $ref resolves to another $ref
    # or a $ref to a $merge:
    #
    #    { $ref: <target1> } --> { $ref: <target2> } --> { <value2> }
    #

    # Minimize copies so that we don't bloat memory
    done = False
    is_copy = False
    while not done:
        if '$merge' in obj:
            with Parser(obj['$merge'], 'eval_shallow') as merge_parser:
                merge_source = merge_parser.parse('source', save=False,
                                                  required=True)
                merge_with = merge_parser.parse('with', save=False,
                                                required=True)

            # This always returns a copy
            obj = json_merge_patch(servicedef, merge_source, merge_with)
            is_copy = True

        elif '$ref' in obj:
            if len(obj.keys()) != 1:
                raise ParseError(
                    "$ref object may not have any other properties", obj)

            sch = servicedef.find(obj['$ref'])
            obj = sch.input
            is_copy = False

        else:
            done = True

    if not is_copy and need_copy:
        obj = copy.copy(obj)

    return obj


def json_merge_patch(servicedef, source, with_):
    """Return a new dict from source with with_ as a json-merge-patch. """

    if isinstance(source, list) or isinstance(with_, list):
        return with_

    if not isinstance(source, dict):
        raise TypeError('source must be a dict, got %s' % (type(source)))

    if not isinstance(with_, dict):
        raise TypeError('with_ must be a dict, got %s' % (type(with_)))

    isdebug() and logger.debug('JSON merge:\nsource = %s\nwith = %s' %
                               (json.dumps(source, indent=2),
                                json.dumps(with_, indent=2)))

    if (  '$ref' in source and
          '$ref' in with_ and
          source == with_):
        # If merging 2 $refs and they have the same target, nothing
        # to do ... but avoid infinite recusion!
        return source

    # Need to make a copy of source, as this is going to be modified
    # Only make a shallow copy here - only the shallow properties
    # are modified in this call.  Recursively deeper calls will make
    # deeper copies as needed.
    source = _eval_shallow(servicedef, source, need_copy=True)

    # with_ is only used in a readonly fashion, so no need to copy
    with_ = _eval_shallow(servicedef, with_, need_copy=False)

    for key, value in with_.iteritems():
        if value is None:
            # Remove the key if present in the source
            if key in source:
                del source[key]
        elif (  isinstance(value, dict) and
                key in source and
                isinstance(source[key], dict)):
            # If this key is a dict in both source and with_, recurse
            source[key] = json_merge_patch(servicedef, source[key], value)
        else:
            # Otherwise update the source for this key.  This may add the
            # key to source if it was not already present
            source[key] = value

    isdebug() and logger.debug('JSON merge result:\n%s' %
                               (json.dumps(source, indent=2)))

    return source
